/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import mx4j.log.Log;
import mx4j.log.Logger;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public class NotificationBroadcasterSupport implements NotificationBroadcaster
{
	private static final NotificationFilter NULL_FILTER = new NotificationFilter()
	{
		public boolean isNotificationEnabled(Notification notification) {return true;}
		public String toString() {return "null filter";}
	};
	private static final Object NULL_HANDBACK = new Object() {public String toString() {return "null handback";}};

	private Map m_filters = new HashMap();
	private Map m_handbacks = new HashMap();
	private Logger m_logger;

	public NotificationBroadcasterSupport() {}

	private Logger getLogger()
	{
		return Log.getLogger("jmx.mbean." + getClass().getName());
	}

	public void addNotificationListener(NotificationListener listener, NotificationFilter filter, Object handback)
	{
		Logger logger = getLogger();

		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Adding notification listener: " + listener + ", filter: " + filter + ", handback: " + handback);}

		if (listener == null) {throw new RuntimeOperationsException(new IllegalArgumentException("Notification listener cannot be null"));}

		if (filter == null) {filter = NULL_FILTER;}
		if (handback == null) {handback = NULL_HANDBACK;}

		synchronized (this)
		{
            List filts = (List)m_filters.get(listener);
			List hands = (List)m_handbacks.get(listener);
			if (filts == null)
			{
				filts = new ArrayList();
				m_filters.put(listener, filts);
				hands = new ArrayList();
				m_handbacks.put(listener, hands);
				filts.add(filter);
				hands.add(handback);
			}
			else
			{
				// Check that is not already registered
				for (int i = 0; i < filts.size(); ++i)
				{
					if (filts.get(i).equals(filter))
					{
						// Same filter, check handback
						if (hands.get(i).equals(handback))
						{
							// Same filter and same handback for the same listener, it's already registered
							throw new RuntimeOperationsException(new IllegalArgumentException("Notification listener is already registered"));
						}
					}
				}
				// Not yet registered, register.
				// Do not merge these 2 calls with the ones above: like this is easier to debug
				// (I know the if-else branch from where I'm coming)
				filts.add(filter);
				hands.add(handback);
			}

			if (logger.isEnabledFor(Logger.DEBUG))
			{
				logger.debug("Filters for this listener: " + filts);
				logger.debug("Handbacks for this listener: " + hands);
			}
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("Notification listener added successfully");}
		}
	}

	public void removeNotificationListener(NotificationListener listener) throws ListenerNotFoundException
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Removing notification listener: " + listener);}

		synchronized (this)
		{
			List filts = (List)m_filters.remove(listener);
			m_handbacks.remove(listener);
			if (filts == null) {throw new ListenerNotFoundException("Notification listener not found");}
			if (logger.isEnabledFor(Logger.INFO)) {logger.info(filts.size() + "notification listener removed successfully");}
		}
	}
/*
	public void removeNotificationListener(NotificationListener listener, Object handback) throws ListenerNotFoundException
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Removing notification listener: " + listener + ", handback: " + handback);}

		if (handback == null) {handback = NULL_HANDBACK;}
		synchronized (this)
		{
			List hands = (List)m_handbacks.get(listener);
			if (hands == null) {throw new ListenerNotFoundException("Notification listener not found");}
			List filts = (List)m_filters.get(listener);

			// Search for the given handback(s)
			int count = 0;
			for (int i = 0; i < hands.size(); ++i)
			{
				Object hand = hands.get(i);
				if (hand.equals(handback))
				{
					hands.remove(i);
					filts.remove(i);
					++count;
				}
			}

			// Check if it was the last listener
			if (hands.isEmpty())
			{
				m_filters.remove(listener);
				m_handbacks.remove(listener);
			}

			if (logger.isEnabledFor(Logger.DEBUG))
			{
				logger.debug("Filters for this listener: " + filts);
				logger.debug("Handbacks for this listener: " + hands);
			}
			if (logger.isEnabledFor(Logger.INFO)) {logger.info(count + "notification listener removed successfully");}
		}
	}

	public void removeNotificationListener(NotificationListener listener, NotificationFilter filter, Object handback) throws ListenerNotFoundException
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Removing notification listener: " + listener + ", filter: " + filter + ", handback: " + handback);}

		if (filter == null) {filter = NULL_FILTER;}
		if (handback == null) {handback = NULL_HANDBACK;}

		synchronized (this)
		{
			List filts = (List)m_filters.get(listener);
			if (filts == null) {throw new ListenerNotFoundException("Notification listener not found");}
			List hands = (List)m_handbacks.get(listener);

			// Search for the given filter and handback(s)
			int count = 0;
			for (int i = 0; i < filts.size(); ++i)
			{
				NotificationFilter filt = (NotificationFilter)filts.get(i);
				if (filt.equals(filter))
				{
					// Same filter, check if same handback
					Object hand = hands.get(i);
					if (hand.equals(handback))
					{
						hands.remove(i);
						filts.remove(i);
						++count;
					}
				}
			}

			// Check if it was the last listener
			if (filts.isEmpty())
			{
				m_filters.remove(listener);
				m_handbacks.remove(listener);
			}

			if (logger.isEnabledFor(Logger.DEBUG))
			{
				logger.debug("Filters for this listener: " + filts);
				logger.debug("Handbacks for this listener: " + hands);
			}
			if (logger.isEnabledFor(Logger.INFO)) {logger.info(count + "notification listener removed successfully");}
		}
	}
*/
	public MBeanNotificationInfo[] getNotificationInfo()
	{
		// Subclasses should override returning more informations
		return new MBeanNotificationInfo[0];
	}

	private boolean hasListeners()
	{
		synchronized (this)
		{
			return m_filters.size() > 0;
		}
	}

	public void sendNotification(final Notification n)
	{
		final Logger logger = getLogger();
		boolean debug = logger.isEnabledFor(Logger.DEBUG);
		final boolean info = logger.isEnabledFor(Logger.INFO);
		synchronized (this)
		{
			// Loop over all listeners
			Iterator i = m_filters.keySet().iterator();

			if (i.hasNext() && info)
			{
				logger.info("Sending notifications...");
			}

			while (i.hasNext())
			{
				final NotificationListener listener = (NotificationListener)i.next();
				if (info) {logger.info("\tListener is: " + listener);}

				List filts = (List)m_filters.get(listener);
				List hands = (List)m_handbacks.get(listener);

				if (debug)
				{
					logger.debug("\tFilters for this listener: " + filts);
					logger.debug("\tHandbacks for this listener: " + hands);
				}

				// Loop over the same listener that registered many times with different filter / handbacks
				for (int j = 0; j < filts.size(); ++j)
				{
					final NotificationFilter filter = (NotificationFilter)filts.get(j);
					Object hb = hands.get(j);
					if (hb == NULL_HANDBACK) {hb = null;}
					final Object handback = hb;

					Runnable run = new Runnable()
					{
						public void run()
						{
							boolean enabled = false;
							try
							{
								enabled = filter != null && filter.isNotificationEnabled(n);
							}
							catch (RuntimeException x)
							{
								if (info) {logger.info("\t\tRuntimeException caught from isNotificationEnabled", x);}
								// And go on
							}

							if (info) {logger.info("\t\tFilter is: " + filter + ", enabled: " + enabled);}

							if (enabled)
							{
								if (info)
								{
									logger.info("\t\tHandback is: " + handback);
									logger.info("\t\tSending notification " + n);
								}
								try
								{
									listener.handleNotification(n, handback);
								}
								catch (RuntimeException x)
								{
									if (info) {logger.info("\t\tRuntimeException caught from handleNotification", x);}
									// And go on with next listener
								}
							}
						}
					};

					if (listener instanceof Serializable)
					{
						// It's remote, run in separate thread
						new Thread(run, "NotificationBroadcasterSupport Thread").start();
					}
					else
					{
						// Run in same thread, it's local
						run.run();
					}
				}
			}
		}
	}
}
