/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.loading;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLStreamHandlerFactory;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.StringTokenizer;

import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.ServiceNotFoundException;

import mx4j.loading.MLetParser;
import mx4j.loading.MLetTag;
import mx4j.loading.MLetParseException;
import mx4j.log.Log;
import mx4j.log.Logger;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @author <a href="mailto:shadow12@users.sourceforge.net">Bronwen Cassidy</a>
 * @version $Revision: 1.10 $
 */
public class MLet extends URLClassLoader implements MLetMBean, MBeanRegistration
{
	private String m_libraryDirectory;
	private ObjectName m_objectName;
	private MBeanServer m_server;
	private String m_defaultCodeBase;
	private boolean m_loadingWithRepository;

	public MLet() {super(new URL[0]);}
	public MLet(URL[] urls) {super(urls);}
	public MLet(URL[] urls, ClassLoader parent) {super(urls, parent);}
	public MLet(URL[] urls, ClassLoader parent, URLStreamHandlerFactory factory) {super(urls, parent, factory);}

	public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception
	{
		m_server = server;
		m_objectName = name == null ? new ObjectName(m_server.getDefaultDomain(), "service", "MLet") : name;
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("MLet service preRegistered");}
		return m_objectName;
	}
	public void postRegister(Boolean registrationDone)
	{
		Logger logger = getLogger();
		boolean done = registrationDone.booleanValue();
		if (!done)
		{
			m_server = null;
			logger.warn("MLet service was NOT registered");
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("MLet service postRegistered");}
		}
	}
	public void preDeregister() throws Exception
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) {logger.debug("MLet service preDeregistered");}
	}
	public void postDeregister()
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) {logger.debug("MLet service postDeregistered");}
	}

	public void addURL(String url) throws ServiceNotFoundException
	{
		addURL(createURL(url));
	}

	public void addURL(URL url)
	{
		Logger logger = getLogger();
        if (!Arrays.asList(getURLs()).contains(url))
		{
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("Adding URL to this ClassLoader classpath: " + url);}
			super.addURL(url);
		}
		else
		{
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("URL " + url + " already present in this ClassLoader classpath");}
		}
	}

	public Class loadClass(String name, ClassLoaderRepository repository) throws ClassNotFoundException
	{
		if (repository == null)
		{
			return loadClass(name);
		}
		else
		{
			try
			{
				Class cls = findClassLocally(name);
				return cls;
			}
			catch (ClassNotFoundException x)
			{
				// Not found locally, try the given repository
				Class cls = findClassInRepository(name, repository);
				return cls;
			}
		}
	}

	protected Class findClass(String name) throws ClassNotFoundException
	{
		if (m_loadingWithRepository) {throw new ClassNotFoundException(name);}

		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Finding class " + name + "...");}
		try
		{
			Class cls = findClassLocally(name);
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("Class " + name + " found in this ClassLoader classpath: " + this);}
			return cls;
		}
		catch (ClassNotFoundException x)
		{
			if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Class " + name + " not found in this ClassLoader classpath: " + this + ", trying the ClassLoaderRepository...");}
			try
			{
				m_loadingWithRepository = true;
				Class cls = findClassInRepository(name, MBeanServerFactory.getClassLoaderRepository(m_server));
				if (logger.isEnabledFor(Logger.INFO)) {logger.info("Class " + name + " found with the ClassLoaderRepository");}
				return cls;
			}
			catch (ClassNotFoundException xx)
			{
				if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Class " + name + " not found in the ClassLoaderRepository, trying the DefaultLoaderRepository", xx);}
				try
				{
					// Try with the static classloader repository
					Class cls = findClassInRepository(name, MBeanServerFactory.getClassLoaderRepository(null));
					if (logger.isEnabledFor(Logger.INFO)) {logger.info("Class " + name + " found with the DefaultLoaderRepository");}
					return cls;
				}
				catch (ClassNotFoundException xxx)
				{
					if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Class " + name + " not found in the DefaultLoaderRepository, giving up", xxx);}
					throw xxx;
				}
			}
			finally
			{
				m_loadingWithRepository = false;
			}
		}
	}

	private Class findClassLocally(String name) throws ClassNotFoundException
	{
		return super.findClass(name);
	}

	private Class findClassInRepository(String name, ClassLoaderRepository repository) throws ClassNotFoundException
	{
		// Try with the classloader repository of the same MBeanServer
		return repository.loadClassWithout(this, name);
	}

	public Set getMBeansFromURL(String url) throws ServiceNotFoundException
	{
		return getMBeansFromURL(createURL(url));
	}

	public Set getMBeansFromURL(URL url) throws ServiceNotFoundException
	{
		if (url == null) {throw new ServiceNotFoundException("Cannot load MBeans from null URL");}

		Logger logger = getLogger();
		String urlString = url.toString();
		m_defaultCodeBase = urlString.substring(0, urlString.lastIndexOf('/') + 1);

		InputStream is = null;
		try
		{
			is = url.openStream();
			if (logger.isEnabledFor(Logger.INFO)) {logger.info("Parsing MLet file " + url);}
			return getMBeansFromURL(is);
		}
		catch (IOException x)
		{
			if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Cannot open MLet URL", x);}
			throw new ServiceNotFoundException(x.toString());
		}
		finally
		{
			if (is != null) {try {is.close();} catch (IOException ignored) {}}
		}
	}

	private Logger getLogger()
	{
		return Log.getLogger("jmx.mbean." + getClass().getName());
	}

	private String getDefaultCodeBase()
	{
		if (m_defaultCodeBase == null)
		{
			String userDir = System.getProperty("user.dir");
			File f = new File(userDir);
			try
			{
				String url = f.toURL().toString();
				m_defaultCodeBase = url.substring(0, url.lastIndexOf('/') + 1);
			}
			catch (MalformedURLException ignored)
			{
				m_defaultCodeBase = "";
			}
		}
		return m_defaultCodeBase;
	}

	private Set getMBeansFromURL(InputStream is) throws ServiceNotFoundException
	{
		Logger logger = getLogger();
		ByteArrayOutputStream baos = new ByteArrayOutputStream();
		BufferedOutputStream os = new BufferedOutputStream(baos);
		try
		{
			readFromAndWriteTo(is, os);
		}
		catch (IOException x)
		{
			logger.error("Cannot read stream", x);
			throw new ServiceNotFoundException(x.toString());
		}
		finally
		{
			try {os.close();} catch (IOException ignored) {}
		}

      try
      {
         HashSet mbeans = new HashSet();
         MLetParser parser = new MLetParser();
         List tags = parser.parse(new String(baos.toByteArray()));

         for (int i = 0; i < tags.size(); ++i)
         {
            MLetTag tag = (MLetTag)tags.get(i);

            // Add the MBean's codebase to the MLet classloader
            URL[] urls = tag.getURLs(getDefaultCodeBase());
            for (int j = 0; j < urls.length; ++j)
            {
               addURL(urls[j]);
            }

            // Create and register the MBean
            Object obj = createMBean(tag);
            mbeans.add(obj);
         }

         return mbeans;
      }
      catch (MLetParseException x)
      {
         throw new ServiceNotFoundException(x.toString());
      }
   }

	protected String findLibrary(String libraryName)
	{
		// When asked to load a library, I should convert the library name to the system specific name
		// For example if libraryName == "stat", on Solaris the conversion yields libstat.so, on W2K yields stat.dll
		String sysLibraryName = System.mapLibraryName(libraryName);

		InputStream is = getResourceAsStream(sysLibraryName.replace(File.separatorChar, '/'));
		// try to load the native libaray from jar only using the library name
        String path = loadResource(sysLibraryName, is);
		if(path != null) return path;

		// library not found so try to load the library using the os specific architecture
        StringBuffer sysBuff = new StringBuffer();
		sysBuff.append(removeWhiteSpace(System.getProperty("os.name").trim())).append(File.separator);
		sysBuff.append(removeWhiteSpace(System.getProperty("os.arch").trim())).append(File.separator);
		sysBuff.append(removeWhiteSpace(System.getProperty("os.version").trim())).append(File.separator);
		sysBuff.append("lib").append(File.separator).append(sysLibraryName);

		path = loadResource(sysBuff.toString(), is);
		if(path != null) return path;

		// 3. Search the java.library.path
		StringTokenizer tokenizer = new StringTokenizer(System.getProperty("java.library.path").trim(), File.pathSeparator);
		while(tokenizer.hasMoreTokens())
		{
			String filePath = new StringBuffer(tokenizer.nextToken().trim()).append(File.separator).append(sysLibraryName).toString();
			try
			{
				path = loadResource(filePath, new FileInputStream(filePath));
				if(path != null) return path;
			}
			catch(FileNotFoundException ex) {}
		}
		// all paths exhausted return null
		return null;
	}

	private String loadResource(String sysLibraryName, InputStream is)
	{
		if (is != null)
		{
			// I should store the library in the given library directory
			try
			{
				File file = null;
				if (getLibraryDirectory() == null)
				{
					file = new File(sysLibraryName);
				}
				else
				{
					file = new File(getLibraryDirectory(), sysLibraryName);
				}
				OutputStream os = null;
				try
				{
					os = new BufferedOutputStream(new FileOutputStream(file));
					is = new BufferedInputStream(is);
					// Delete if already exists
					if (file.exists()) {file.delete();}
					readFromAndWriteTo(is, os);
					return file.getCanonicalPath();
				}
				finally
				{
					if (is != null) {is.close();}
					if (os != null) {os.close();}
				}
			}
			catch (IOException ignored) {}
		}
		return null;
	}

	private void readFromAndWriteTo(InputStream is, OutputStream os) throws IOException
	{
		byte[] buffer = new byte[65536];
		int read = -1;
		while ((read = is.read(buffer)) >= 0)
		{
			os.write(buffer, 0, read);
		}
		buffer = null;
	}

	private String removeWhiteSpace(String word)
	{
		int j = word.indexOf(' ');
		// no spaces in the word return it
		if(j == -1) return word;

		StringBuffer temp = new StringBuffer();
		int i = 0;
		while(j != -1)
		{
			word = word.substring(i);
			j = word.indexOf(' ');
			// store the word up to the space in temp
			if(j != -1) { temp = temp.append(word.substring(0, j)); }
			else { temp = temp.append(word.substring(0));}
			i = j + 1;
		}
		return temp.toString();
	}

	public String getLibraryDirectory() {return m_libraryDirectory;}
	public void setLibraryDirectory(String libdir) {m_libraryDirectory = libdir;}

	private Object createMBean(MLetTag tag) throws ServiceNotFoundException
	{
		if (m_server == null) {throw new ServiceNotFoundException("MLet not registered on the MBeanServer");}

		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.INFO)) {logger.info("Creating MBean from\n" + tag);}

		try
		{
			Object mbean = null;
         if (tag.getObject() != null)
			{
				// Read the file from the codebase URLs of this classloader
				String name = tag.getObject();
				InputStream is = getResourceAsStream(name);
				if (is == null) {throw new ServiceNotFoundException("Cannot find MBean " + name + " in this ClassLoader classpath");}

				InputStream bis = new BufferedInputStream(is);
				ByteArrayOutputStream baos = new ByteArrayOutputStream();
				OutputStream os = new BufferedOutputStream(baos);
				readFromAndWriteTo(bis, os);

				// Deserialize using the MLet classloader
				ObjectInputStream ois = m_server.deserialize(m_objectName, baos.toByteArray());
				mbean = ois.readObject();
			}
			else
			{
				// Instantiate using the MLet classloader
				String clsName = tag.getCode();
				Object[] args = tag.getArguments();
				String[] params = tag.getSignature();
				mbean = m_server.instantiate(clsName, m_objectName, args, params);
			}

			ObjectName objectName = tag.getObjectName();
			ObjectInstance instance = m_server.registerMBean(mbean, objectName);
			return instance;
		}
		catch (Throwable t)
		{
			return t;
		}
	}

	private URL createURL(String urlString) throws ServiceNotFoundException
	{
		try
		{
			URL url = new URL(urlString);
			return url;
		}
		catch (MalformedURLException x)
		{
			throw new ServiceNotFoundException(x.toString());
		}
	}
}
