/**
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package javax.management.relation;

import javax.management.NotificationBroadcasterSupport;
import javax.management.ObjectName;
import javax.management.MBeanServer;
import javax.management.MBeanRegistration;
import javax.management.NotificationListener;
import javax.management.MBeanServerNotification;
import javax.management.InstanceNotFoundException;
import javax.management.Notification;
import javax.management.NotificationFilter;
import javax.management.JMException;
import javax.management.MBeanException;
import javax.management.ReflectionException;
import javax.management.Attribute;
import javax.management.AttributeNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanNotificationInfo;
import javax.management.RuntimeOperationsException;
import javax.management.MalformedObjectNameException;

import mx4j.log.Logger;
import mx4j.log.Log;
import mx4j.util.StandardMBeanProxy;

import java.util.Map;
import java.util.HashMap;
import java.util.List;
import java.util.ArrayList;
import java.util.Set;
import java.util.Iterator;
import java.util.Vector;

import java.util.Date;

/**
 * @author <a href="mailto:shadow12@users.sourceforge.net">Bronwen Cassidy</a>
 * @version $Revision: 1.8 $
 */
public class RelationService extends NotificationBroadcasterSupport implements RelationServiceMBean,
			                 MBeanRegistration, NotificationListener
{
	private boolean m_purgeFlag;
	private Long m_notificationCounter = new Long(0);

	private MBeanServer m_server = null;
	private RelationSupportMBean m_proxy = null;
	private InternalRelation m_internalRelation = null;
	private ObjectName m_relationServiceObjectName = null;

	private MBeanServerNotificationFilter m_notificationFilter = null;

	private Map m_relationIdToRelationObject = new HashMap();
	private Map m_relationIdToRelationTypeName = new HashMap();
	private Map m_relationMBeanObjectNameToRelationId = new HashMap();
	private Map m_relationTypeNameToRelationTypeObject = new HashMap();
	private Map m_relationTypeNameToRelationIds = new HashMap();
	private Map m_referencedMBeanObjectNameToRelationIds = new HashMap();

	private List m_deregisteredNotificationList = new ArrayList();

	public RelationService(boolean purgeFlag)
	{
		m_purgeFlag = purgeFlag;
	}

	public void isActive() throws RelationServiceNotRegisteredException
	{
		Logger logger = getLogger();
		if(m_server == null)
		{
			logger.error("RelationService has not been registered in the MBeanServer");
			throw new RelationServiceNotRegisteredException("Relation Service is not registered");
		}
	}

	public boolean getPurgeFlag()
	{
		return m_purgeFlag;
	}

	public void setPurgeFlag(boolean purgeFlag)
	{
		m_purgeFlag = purgeFlag;
	}

	public void createRelationType(String relationTypeName, RoleInfo[] roleInfos) throws IllegalArgumentException,
										  InvalidRelationTypeException
	{
		if(relationTypeName == null) throw new IllegalArgumentException("Illegal Null Relation Type Name value");
		if(roleInfos == null) throw new IllegalArgumentException("Illegal Null RoleInfo");

		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO)) logger.info("Creating Relation Type with relationTypeName: " + relationTypeName);

		RelationTypeSupport relationType = new RelationTypeSupport(relationTypeName, roleInfos);
		// created a new RelationType add it to our map
		addRelationTypeToMap(relationTypeName, relationType);
	}

	private void addRelationTypeToMap(String relationTypeName, RelationType relationType)
	{
		Logger logger = getLogger();
		// synchronize all activities to map.
		synchronized(m_relationTypeNameToRelationTypeObject)
		{
			if(((RelationType)(m_relationTypeNameToRelationTypeObject.get(relationTypeName))) != null)
			{
				logger.error("Cannot addRelationType as a relationType of the same name: " + relationTypeName + " already exists in the RelationService");
				throw new IllegalArgumentException("RelationType with name: " + relationTypeName + " already exists in the RelationService");
			}
			// set the RelationTypeSupport internal flag to true indicating that the relationType has been declared in the relation service
			if(relationType instanceof RelationTypeSupport)
			{
				((RelationTypeSupport)relationType).setRelationServiceFlag(true);
			}
			// store the instance in the map
		    m_relationTypeNameToRelationTypeObject.put(relationTypeName, relationType);
		}
	}

	private RelationType getRelationType(String relationTypeName)
			throws IllegalArgumentException, RelationTypeNotFoundException
	{
		if(relationTypeName == null) throw new IllegalArgumentException("Null relation type name");
		RelationType relationType = null;
		synchronized(m_relationTypeNameToRelationTypeObject)
		{
			relationType = ((RelationType)(m_relationTypeNameToRelationTypeObject.get(relationTypeName)));
			// exception is caught by calling classes testing for RelationTypeNotFound.
			if(relationType == null)
		    {
			    throw new RelationTypeNotFoundException("No RelationType found for relationTypeName: " + relationTypeName);
		    }
			return relationType;
		}
	}

	public void addRelationType(RelationType relationType) throws IllegalArgumentException,
		                                   InvalidRelationTypeException
	{
		if(relationType == null) throw new IllegalArgumentException("Relation Type should not be null.");

		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO)) logger.info("Adding a RelationType");

		List roleInfoList = relationType.getRoleInfos();
		if(roleInfoList == null)
		{
			logger.error("Cannot add RelationType: " + relationType.getClass().getName() + " RoleInfo information was not provided with the RelationType.");
			throw new IllegalArgumentException("No RoleInfo provided with Relation Type");
		}
		// build the roleInfo[] to validate the RoleInfo
		RoleInfo[] roleInfos = new RoleInfo[roleInfoList.size()];
		int index = 0;
		for(Iterator i = roleInfoList.iterator(); i.hasNext();)
		{
			RoleInfo currentRoleInfo = (RoleInfo)i.next();
			roleInfos[index] = currentRoleInfo;
			index++;
		}
		// need to validateRoleInfos before adding
		RelationTypeSupport.checkRoleInfos(roleInfos);
		String relationTypeName = relationType.getRelationTypeName();
		// validated add the RelationType
		addRelationTypeToMap(relationTypeName, relationType);
	}

	public List getAllRelationTypeNames()
	{
		List result = null;
		synchronized(m_relationTypeNameToRelationTypeObject)
		{
			result = new ArrayList(m_relationTypeNameToRelationTypeObject.keySet());
		}
		// return all relationTypeNames or an empty list if none
		return result;
	}

	public List getRoleInfos(String relationTypeName) throws IllegalArgumentException,
		                                   RelationTypeNotFoundException
	{
		if(relationTypeName == null) throw new IllegalArgumentException("Illegal relationType name is null.");
		RelationType relationType = getRelationType(relationTypeName);
		// returns a List of RoleInfo objects or an empty list if none
		return relationType.getRoleInfos();
	}

	public RoleInfo getRoleInfo(String relationTypeName, String roleInfoName) throws IllegalArgumentException,
									       RelationTypeNotFoundException, RoleInfoNotFoundException
	{
		if(relationTypeName == null) throw new IllegalArgumentException("Null relation type name");
		if(roleInfoName == null)throw new IllegalArgumentException("Null RoleInfo name");
		// gets the RelationType then gets the RoleInfo object
		return (getRelationType(relationTypeName).getRoleInfo(roleInfoName));
	}

	public void removeRelationType(String relationTypeName) throws IllegalArgumentException,
		                                   RelationServiceNotRegisteredException, RelationTypeNotFoundException
	{
		Logger logger = getLogger();
		isActive();
		if(relationTypeName == null) throw new IllegalArgumentException("Illegal: relationType name cannot be null.");
		if(logger.isEnabledFor(Logger.INFO)) logger.info("Removing RelationType with relationTypeName: " + relationTypeName);

		RelationType relationType = getRelationType(relationTypeName);
		List relationIdList = null;
		List tempList = getRelationIds(relationTypeName);

		// no need to clone as relationIdList is internal and get its values from a private method.
		if(tempList != null) relationIdList = tempList;
		removeRelationTypeObject(relationTypeName);
		removeRelationTypeName(relationTypeName);
		// we need to remove more references as Relation is referenced in more than one relationId
		if(relationIdList != null)
		{
			for(Iterator i = relationIdList.iterator(); i.hasNext();)
			{
				String currentRelationId = (String)i.next();
				try
				{
					// removed the relationType now remove the relation
					removeRelation(currentRelationId);
				}
				catch(RelationNotFoundException ex) {throw new RuntimeOperationsException(ex.getMessage());}
			}
		}
	}

	private void removeRelationTypeObject(String relationTypeName)
	{
		synchronized(m_relationTypeNameToRelationTypeObject)
		{
		    m_relationTypeNameToRelationTypeObject.remove(relationTypeName);
		}
	}

	private List getRelationIds(String relationTypeName)
	{
		synchronized(m_relationTypeNameToRelationIds)
		{
		    return ((List)(m_relationTypeNameToRelationIds.get(relationTypeName)));
		}
	}

	/**
	 * According to the RI spec this method is used only to create internal relations
	 * so can instantiate the InternalRelation class here and isRelationMBean() should return null.
	 */
	public void createRelation(String relationId, String relationTypeName, RoleList roleList) throws IllegalArgumentException,
		                                   RelationServiceNotRegisteredException, RoleNotFoundException,
										   InvalidRelationIdException, RelationTypeNotFoundException, InvalidRoleValueException
	{
		isActive();
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		if(relationTypeName == null)throw new IllegalArgumentException("Null Relation Type Name");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))
		{
			logger.info("Creating an InternalRelation with ID: " + relationId + " and relationType name: " + relationTypeName);
		}
		// creating InternalRelation to represent the internal relations
		m_internalRelation = new InternalRelation(relationId, m_relationServiceObjectName,
						relationTypeName, roleList);
		try
		{
			if(getRelationObject(relationId) != null)
			{
				logger.error("There is a Relation already registered in the RelationServcie with ID: " + relationId);
				throw new InvalidRelationIdException("There is already a relation with id: " + relationId);
			}
		}
		catch (RelationNotFoundException ex) {/*Do nothing as should not be found*/}

		RelationType relationType = getRelationType(relationTypeName);
		ArrayList roleInfoList = (ArrayList)(buildRoleInfoList(relationType, roleList));
		if(!(roleInfoList.isEmpty()))
		{
			initializeMissingCreateRoles(roleInfoList, m_internalRelation, relationId, relationTypeName);
		}

		addRelationObject(relationId, m_internalRelation);
		addRelationId(relationId, relationTypeName);
		addRelationTypeName(relationId, relationTypeName);
		updateRoles(roleList, relationId);
		try
		{
			if(logger.isEnabledFor(Logger.INFO)) logger.info("sending RelationCreation notification to all listeners");
			sendRelationCreationNotification(relationId);
		}
		catch (RelationNotFoundException ex) { throw new RuntimeOperationsException("Unable to send notification as Relation not found");}
	}

	private void updateRoles(RoleList roleList, String relationId) throws RelationServiceNotRegisteredException,
			IllegalArgumentException
	{
		if(roleList == null) throw new IllegalArgumentException("Null RoleList");
		if(relationId == null) throw new IllegalArgumentException("Null relationId");
		for(Iterator i = roleList.iterator(); i.hasNext();)
		{
			Role currentRole = (Role)i.next();
			ArrayList tempList = new ArrayList();
			try
			{
			    updateRoleMap(relationId, currentRole, tempList);
			}
			catch (RelationNotFoundException ex) {throw new RuntimeOperationsException("Cannot update the roleMap as Relation not found");}
		}
	}

	private List buildRoleInfoList(RelationType relationType, List roleList) throws InvalidRoleValueException,
			RoleNotFoundException
	{
		List roleInfoList = relationType.getRoleInfos();
		if(roleList != null)
		{
			for(Iterator i = roleList.iterator(); i.hasNext();)
			{
				Role currentRole = (Role)i.next();
				String currentRoleName = currentRole.getRoleName();
				List currentRoleValue = currentRole.getRoleValue();
				RoleInfo roleInfo = null;
				try
				{
					roleInfo = relationType.getRoleInfo(currentRoleName);
				}
				catch (RoleInfoNotFoundException ex)
				{
					throw new RoleNotFoundException(ex.getMessage());
				}
				int problemType = (checkRoleCardinality(currentRoleName, currentRoleValue, roleInfo)).intValue();
				if(problemType != 0)
				{
					throwRoleProblemException(problemType, currentRoleName);
				}
				roleInfoList.remove(roleInfoList.indexOf(roleInfo));
			}
		}
		return roleInfoList;
	}

	private void addRelationTypeName(String relationId, String relationTypeName)
	{
		synchronized(m_relationTypeNameToRelationIds)
		{
			ArrayList idList = (ArrayList)m_relationTypeNameToRelationIds.get(relationTypeName);
			boolean isNewRelation = false;
			if(idList == null)
			{
				isNewRelation = true;
				idList = new ArrayList();
			}
			idList.add(relationId);
			if(isNewRelation) m_relationTypeNameToRelationIds.put(relationTypeName, idList);
		}
	}

	private void addRelationObject(String relationId, RelationSupport relationSupport)
	{
		synchronized(m_relationIdToRelationObject)
		{
			m_relationIdToRelationObject.put(relationId, relationSupport);
		}
	}

	private void addRelationObjectName(String relationId, ObjectName relationObjectName)
	{
		synchronized(m_relationIdToRelationObject)
		{
			m_relationIdToRelationObject.put(relationId, relationObjectName);
		}
	}

	private void addRelationId(String relationId, String relationTypeName)
	{
		synchronized(m_relationIdToRelationTypeName)
		{
			m_relationIdToRelationTypeName.put(relationId, relationTypeName);
		}
	}

	// method gets called only for internal relations
	private void initializeMissingCreateRoles(List roleInfoList, InternalRelation internalRelation,
			String relationId, String relationTypeName)
			throws RelationTypeNotFoundException, RelationServiceNotRegisteredException,
			InvalidRoleValueException, RoleNotFoundException, IllegalArgumentException
	{
		isActive();
		if(roleInfoList == null) throw new IllegalArgumentException("RoleInfoList is Null");
		if(relationId == null) throw new IllegalArgumentException("RelationId is Null.");
		if(relationTypeName == null) throw new IllegalArgumentException("Relation Type Name is Null.");

		for(Iterator i = roleInfoList.iterator(); i.hasNext();)
	    {
		    RoleInfo currentRoleInfo = (RoleInfo)i.next();
		    String roleName = currentRoleInfo.getName();

			ArrayList temp = new ArrayList();
		    Role role = new Role(roleName, temp);
			try
			{
			    internalRelation.setRole(role);
			}
			catch(RelationNotFoundException ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
		}
	}

	private Integer checkRoleCardinality(String roleName, List roleValue, RoleInfo roleInfo)
	{
		if(roleName == null) throw new IllegalArgumentException("Null Role Name");
		if(roleValue == null) throw new IllegalArgumentException("Null roleValue List");
		if(roleInfo == null) throw new IllegalArgumentException("Null RoleInfo");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO)) logger.info("checking role cardinality");

		if(!(roleName.equals(roleInfo.getName())))
		{
			logger.error("Role does not have a valid roleName");
			return new Integer(RoleStatus.NO_ROLE_WITH_NAME);
		}
		if(!(roleInfo.checkMinDegree(roleValue.size())))
		{
			logger.error("Minimum number of references defined in the RoleInfo has fallen below minimum");
			return (new Integer(RoleStatus.LESS_THAN_MIN_ROLE_DEGREE));
		}
		if(!(roleInfo.checkMaxDegree(roleValue.size())))
		{
			logger.error("Maximum number of references defined in the RoleInfo has gone above the maximum");
			return new Integer(RoleStatus.MORE_THAN_MAX_ROLE_DEGREE);
		}

		String referencedClassName = roleInfo.getRefMBeanClassName();
		for(Iterator i = roleValue.iterator(); i.hasNext();)
		{
			ObjectName currentObjectName = (ObjectName)i.next();
			if(currentObjectName == null)
			{
				logger.error("The mbean with ObjectName: " + currentObjectName.getCanonicalName() + " is not registered in the MBeanServer");
				return new Integer(RoleStatus.REF_MBEAN_NOT_REGISTERED);
			}
			if(!(m_server.isRegistered(currentObjectName)))
			{
				logger.error("The mbean with ObjectName: " + currentObjectName.getCanonicalName() + " is not registered in the MBeanServer");
				return new Integer(RoleStatus.REF_MBEAN_NOT_REGISTERED);
			}
			try
			{
				if(!(m_server.isInstanceOf(currentObjectName, referencedClassName)))
				{
					logger.error("The class referenced: " + currentObjectName.toString() + " does not match the class expected: " +
									  referencedClassName + " in RoleInfo: " + roleInfo.toString());
					return new Integer(RoleStatus.REF_MBEAN_OF_INCORRECT_CLASS);
				}
			}
			catch (InstanceNotFoundException ex)
			{
				return null;
			}
		}
		return new Integer(0);
	}

	/**
	 * Add relation of a class that extends RelationSupport, is an external relation and can only be created with this method
	 * createRelation() is for internal relations only
	 */
	public void addRelation(ObjectName relationMBeanObjectName)throws IllegalArgumentException, RelationServiceNotRegisteredException,
	                                       NoSuchMethodException, InvalidRelationIdException, InstanceNotFoundException,
										   InvalidRelationServiceException, RelationTypeNotFoundException,
										   RoleNotFoundException, InvalidRoleValueException
	{
		isActive();
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("adding a Relation with ObjectName: " + relationMBeanObjectName.toString());
		// checks through the MBeanServer that the class is an instance of the Relation interface which
		// RelationSupport implements
		checkValidRelation(relationMBeanObjectName);
		//create the proxy
		if(m_proxy == null)m_proxy = (RelationSupportMBean)StandardMBeanProxy.create(RelationSupportMBean.class, m_server, relationMBeanObjectName);
		// get the relationId of the class through the MBeanServer
		String relationId = m_proxy.getRelationId();    //getRelationIdAttribute(relationMBeanObjectName);
		if(relationId == null) throw new InvalidRelationIdException("No RelationId provided");
		// obtains the objectname of the RelationService defined by the user

		ObjectName relationServiceObjectName = m_proxy.getRelationServiceName();
		// checks that the RelationService objectName is running in the server
		if(!(checkRelationServiceIsCurrent(relationMBeanObjectName, relationServiceObjectName)))
		{
			throw new InvalidRelationServiceException("The Relation Service referenced in the MBean is not the current one.");
		}
		// get the relationTypeName through the server for the user defined RelationSupport subclass
		String relationTypeName = m_proxy.getRelationTypeName();
		if(relationTypeName == null) throw new RelationTypeNotFoundException("RelationType not found");
		// get the roles
		RoleList roleList = m_proxy.retrieveAllRoles();
		try
		{
			// already have defined relation registered cannot add another with the same id
		    if(getRelationObject(relationId) != null) throw new InvalidRelationIdException("Relation with ID " + relationId + " already exists");
		}
		catch(RelationNotFoundException ex) {/*Do nothing should not be found*/}

		RelationType relationType = getRelationType(relationTypeName);
		ArrayList roleInfoList = (ArrayList)buildRoleInfoList(relationType, roleList);
		if(!(roleInfoList.isEmpty()))
		{
			for(Iterator i = roleInfoList.iterator(); i.hasNext();)
			{
				RoleInfo currentRoleInfo = (RoleInfo)i.next();
				String currentRoleName = currentRoleInfo.getName();
				ArrayList emptyValueList = new ArrayList();
				Role role = new Role(currentRoleName, emptyValueList);
				try
				{
				    m_proxy.setRole(role);
				}
				catch(RelationNotFoundException ex){throw new RuntimeOperationsException(ex.getMessage());}
			}
		}
		// add all info to our corresponding maps
		updateAllInternals(relationId, relationMBeanObjectName, relationTypeName, roleList);
	}

	private void updateAllInternals(String relationId, ObjectName relationMBeanObjectName, String relationTypeName,
			RoleList roleList) throws RelationServiceNotRegisteredException
	{
		// adds key -> relationId value -> relationMBeanObjectName to HashMap
		addRelationObjectName(relationId, relationMBeanObjectName);
		addRelationId(relationId, relationTypeName);
		addRelationTypeName(relationId, relationTypeName);
		updateRoles(roleList, relationId);
		try
		{
			sendRelationCreationNotification(relationId);
		}
		catch (RelationNotFoundException ex) {throw new RuntimeOperationsException("Cannot send a notification for relationId " + relationId + " as relation not found.");}

		addMBeanObjectNameToRelationId(relationMBeanObjectName, relationId);
		m_proxy.setRelationServiceManagementFlag(new Boolean(true));
		List newReferenceList = new ArrayList();
	    newReferenceList.add(relationMBeanObjectName);
	    updateUnregistrationListener(newReferenceList, null);
	}

	private boolean checkRelationServiceIsCurrent(ObjectName relationMBeanObjectName, ObjectName relationServiceObjectName)
	{
		if(relationServiceObjectName == null) return false;
		if(!(relationServiceObjectName.equals(m_relationServiceObjectName))) return false;
		return true;
	}

	private void checkValidRelation(ObjectName relationMBeanObjectName) throws IllegalArgumentException, InstanceNotFoundException
	{
		if(relationMBeanObjectName == null) throw new IllegalArgumentException("Cannot have a null Relation ObjectName");
		Logger logger = getLogger();
		if(!(m_server.isInstanceOf(relationMBeanObjectName, "javax.management.relation.Relation")))
		{
			logger.error("An MBean which is to be added as a Relation must implement the Relation interface");
			throw new InstanceNotFoundException("MBean does implement the Relation interface");
		}
	}

	public ObjectName isRelationMBean(String relationId) throws IllegalArgumentException, RelationNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id.");
		Object result = getRelationObject(relationId);
		if(result instanceof ObjectName)
		{
			return ((ObjectName) result);
		}
		return null;
	}

	public String isRelation(ObjectName objectName) throws IllegalArgumentException
	{
		if(objectName == null) throw new IllegalArgumentException("Null ObjectName");
		return (getMBeanObjectName(objectName));
	}

	public Boolean hasRelation(String relationId) throws IllegalArgumentException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		try
		{
		    Object result = getRelationObject(relationId);
			return new Boolean(true);
		}
		catch (RelationNotFoundException ex)
		{
			return new Boolean(false);
		}
	}

	public List getAllRelationIds()
	{
		synchronized(m_relationIdToRelationObject)
		{
			return (new ArrayList(m_relationIdToRelationObject.keySet()));
		}
	}

	public Integer checkRoleReading(String roleName, String relationTypeName) throws IllegalArgumentException,
		                                   RelationTypeNotFoundException
	{
		if(roleName == null) throw new IllegalArgumentException("Null RoleName");
		if(relationTypeName == null) throw new IllegalArgumentException("Null RelationType name.");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO)) logger.info("checking if Role with RoleName: " + roleName + " is readable");
		RelationType relationType = getRelationType(relationTypeName);
		try
		{
			RoleInfo roleInfo = relationType.getRoleInfo(roleName);
			if(!(roleName.equals(roleInfo.getName()))) return (new Integer(RoleStatus.NO_ROLE_WITH_NAME));
			if(!(roleInfo.isReadable()))
			{
				logger.error("RoleInfo: " + roleInfo.toString() + " cannot be read");
				return (new Integer(RoleStatus.ROLE_NOT_READABLE));
			}
		}
		catch(RoleInfoNotFoundException ex)
		{
			logger.error("roleInfo for roleName: " + roleName + " has not been found.");
			return (new Integer(RoleStatus.NO_ROLE_WITH_NAME));
		}
		return new Integer(0);
	}

	public Integer checkRoleWriting(Role role, String relationTypeName, Boolean isInitialized) throws
										   IllegalArgumentException, RelationTypeNotFoundException
	{
		if(role == null) throw new IllegalArgumentException("checkRoleWriting was given a null Role");
		if(relationTypeName == null) throw new IllegalArgumentException("checkRoleWriting was given a null RelationTypeName");
		if(isInitialized == null)throw new IllegalArgumentException("checkRoleWriting was given a null Boolean");
		Logger logger = getLogger();
		RelationType relationType = getRelationType(relationTypeName);
		String roleName = role.getRoleName();
		if(logger.isEnabledFor(Logger.INFO)) logger.info("checking if Role with RoleName: " + roleName + " is readable");
		ArrayList roleValue = (ArrayList)role.getRoleValue();
		boolean canWrite = true;
		if(isInitialized.booleanValue()) canWrite = false;
		RoleInfo roleInfo = null;
		try
		{
			roleInfo = relationType.getRoleInfo(roleName);
		}
		catch (RoleInfoNotFoundException ex)
		{
			logger.error("roleInfo for roleName: " + roleName + " has not been found.");
			return new Integer(RoleStatus.NO_ROLE_WITH_NAME);
		}
		if(canWrite)
		{
			if(!(roleInfo.isWritable()))
			{
				logger.error("RoleInfo: " + roleInfo.toString() + " cannot be written to.");
				return new Integer(RoleStatus.ROLE_NOT_WRITABLE);
			}
		}
		return (checkRoleCardinality(roleName, roleValue, roleInfo));
	}

	public void sendRelationCreationNotification(String relationId) throws IllegalArgumentException,
		   RelationNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id.");
		Logger logger = getLogger();
		String message = "Creation of relation " + relationId;
		String relationTypeName = getRelationTypeNameFromMap(relationId);

		if(logger.isEnabledFor(Logger.INFO)) logger.info("A relation has been created with ID: " + relationId +
						" and relationTypeName: " + relationTypeName + " ..sending notification");

		ObjectName relationObjectName = isRelationMBean(relationId);
		String notificationType = getCreationNotificationType(relationObjectName);
		long sequenceNumber = getNotificationSequenceNumber().longValue();
		Date currentDate = new Date();
		long timestamp = currentDate.getTime();
		RelationNotification relationNotification = new RelationNotification(notificationType, this, sequenceNumber,
							 timestamp, message, relationId, relationTypeName, relationObjectName, null);
	}

	private Long getNotificationSequenceNumber()
	{
		Long result = null;
	    synchronized(m_notificationCounter)
		{
	        result = new Long(m_notificationCounter.longValue() + 1);
	        m_notificationCounter = new Long(result.longValue());
		}
	    return result;
	}

	private String getCreationNotificationType(ObjectName relationObjectName)
	{
		if(relationObjectName != null)
		{
			return RelationNotification.RELATION_MBEAN_CREATION;
		}
		return RelationNotification.RELATION_BASIC_CREATION;
	}

	public void sendRoleUpdateNotification(String relationId, Role newRole, List oldRoleValues) throws IllegalArgumentException,
		                                   RelationNotFoundException
	{
		if(relationId == null)throw new IllegalArgumentException("Null RelationId");
		if(newRole == null)throw new IllegalArgumentException("Null Role");
		if(oldRoleValues == null)throw new IllegalArgumentException("Null List of role values");

		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("Sending a roleUpdateNotification of Relation with ID: " + relationId);
		String roleName = newRole.getRoleName();
		List newRoleValues = newRole.getRoleValue();
		String newRoleValueMessage = Role.roleValueToString(newRoleValues);
		String oldRoleValueMessage = Role.roleValueToString(oldRoleValues);
		StringBuffer message = new StringBuffer("Value of the role ");
		message.append(roleName);
		message.append(" has changed\nOld value:\n");
		message.append(oldRoleValueMessage);
		message.append("\nNew value:\n");
		message.append(newRoleValueMessage);
		if(logger.isEnabledFor(Logger.INFO))logger.info("Notification message: " + message.toString());
		String relationTypeName = getRelationTypeNameFromMap(relationId);

		// determine if this is a relation update or a relation mbean update
		ObjectName relationObjectName = isRelationMBean(relationId);
		String notificationType = null;
		if(relationObjectName != null) notificationType = RelationNotification.RELATION_MBEAN_UPDATE;
		else notificationType = RelationNotification.RELATION_BASIC_UPDATE;

		long sequenceNumber = getNotificationSequenceNumber().longValue();
		Date currentDate = new Date();
		long timeStamp = currentDate.getTime();

		RelationNotification notification = new RelationNotification(notificationType, this, sequenceNumber, timeStamp,
							 message.toString(), relationId, relationTypeName, relationObjectName, roleName, oldRoleValues,
							 newRoleValues);
		sendNotification(notification);
	}

	public void sendRelationRemovalNotification(String relationId, List unregisteredMBeanList) throws IllegalArgumentException,
		                                   RelationNotFoundException
	{
		if(relationId == null)throw new IllegalArgumentException("Null RelationId");

		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO)) logger.info("sending relationRemovalNotification of ID: " + relationId);
		StringBuffer message = new StringBuffer("Removal of relation ");
		message.append(relationId);
		String relationTypeName = getRelationTypeNameFromMap(relationId);
		ObjectName relationObjectName = isRelationMBean(relationId);
		String notificationType = null;
		if(relationObjectName != null) notificationType = RelationNotification.RELATION_MBEAN_REMOVAL;
		else notificationType = RelationNotification.RELATION_BASIC_REMOVAL;
		long sequenceNumber = getNotificationSequenceNumber().longValue();
		Date currentDate = new Date();
		long timeStamp = currentDate.getTime();
		RelationNotification notification = new RelationNotification(notificationType, this, sequenceNumber, timeStamp,
							 message.toString(), relationId, relationTypeName, relationObjectName, unregisteredMBeanList);

		sendNotification(notification);
	}

	/**
	 * Handles update of the RelationService role map for the update of given roles in a given relation
	 */
	public void updateRoleMap(String relationId, Role role, List oldRoleValues) throws IllegalArgumentException,
		                                   RelationServiceNotRegisteredException, RelationNotFoundException
	{
		isActive();
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		if(role == null) throw new IllegalArgumentException("Null Role");
		if(oldRoleValues == null) throw new IllegalArgumentException("Null Role value list.");

		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO)) logger.info("Updating the RelationService RoleMap");

		Object result = getRelationObject(relationId);
		String roleName = role.getRoleName();
		List newRoleValue = role.getRoleValue();

		// clone as the list is to be modified later, cast to ArrayList as List does not define clone()
		List oldValues = (ArrayList) (((ArrayList)oldRoleValues).clone());

		// List of ObjectNames of new referenced MBeans
		List newReferenceList = new ArrayList();
		for(Iterator i = newRoleValue.iterator(); i.hasNext();)
		{
			ObjectName currentObjectName = (ObjectName)i.next();
			// Check if this ObjectName was already present in oldValueList
			int currentObjectNamePosition = oldValues.indexOf(currentObjectName);
			// we have a new Reference
			if(currentObjectNamePosition == -1)
			{
				// returns true if we have a new reference, false if the MBean is already referenced
				if(addNewMBeanReference(currentObjectName, relationId, roleName))
				{
					// add to new references list
					newReferenceList.add(currentObjectName);
				}
			}
			else
			{
				// MBean referenced in an old value remove
				oldValues.remove(currentObjectNamePosition);
			}
		}
		List obsoleteReferenceList = getObsoleteReferenceList(oldValues, relationId, roleName);
		// update listeners as to the new references
		updateUnregistrationListener(newReferenceList, obsoleteReferenceList);
	}

	private List getObsoleteReferenceList(List oldValues, String relationId, String roleName)
		   throws IllegalArgumentException
	{
		List obsoleteReferenceList = new ArrayList();
		for(Iterator i = oldValues.iterator(); i.hasNext();)
		{
			ObjectName currentObjectName = (ObjectName)i.next();
			if(removeMBeanReference(currentObjectName, relationId, roleName))
			{
				obsoleteReferenceList.add(currentObjectName);
			}
		}
		return obsoleteReferenceList;
	}

	private boolean removeMBeanReference(ObjectName objectName, String relationId, String roleName)
			throws IllegalArgumentException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		if(objectName == null) throw new IllegalArgumentException("Null ObjectName");
		if(roleName == null) throw new IllegalArgumentException("Null Role Name.");
		// first check if we have any references for MBean
		HashMap mbeanReferenceMap = (HashMap)getReferencedMBeansFromMap(objectName);
		if(mbeanReferenceMap == null) return true;
		// we have references get the roleNames for the relationId
		ArrayList roleNames = (ArrayList)(mbeanReferenceMap.get(relationId));
		// roleName found remove it
		if(roleNames.indexOf(roleName) != -1) roleNames.remove(roleNames.indexOf(roleName));
		// no more references remove the relationId key
		if(roleNames.isEmpty()) mbeanReferenceMap.remove(relationId);
		if (mbeanReferenceMap.isEmpty())
		{
			// now we can remove the MBean ObejctName with all references removed
			removeObjectName(objectName);
			return true;
		}
		return false;
	}

	private Map getReferencedMBeansFromMap(ObjectName objectName)
	{
		synchronized(m_referencedMBeanObjectNameToRelationIds)
		{
			return ((HashMap)m_referencedMBeanObjectNameToRelationIds.get(objectName));
		}
	}

	private boolean addNewMBeanReference(ObjectName objectName, String relationId, String roleName)
			throws IllegalArgumentException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		if(roleName == null) throw new IllegalArgumentException("Null Role Name");
		if(objectName == null) throw new IllegalArgumentException("Null ObjectName.");

		// get the ObjectNames of the referenced ObjectName
		HashMap mbeanReferenceMap = (HashMap)getAllMBeanObjectNames(objectName);
		// it is null, add it to our map and return true - we have a new reference
		if(mbeanReferenceMap == null)
		{
			ArrayList roleNames = new ArrayList();
			roleNames.add(roleName);
			mbeanReferenceMap = new HashMap();
			mbeanReferenceMap.put(relationId, roleNames);
			addObjectNameToMBeanReference(objectName, mbeanReferenceMap);
			return true;
		}
		// not a new MBeanReference return false
		return false;
	}

	private void addObjectNameToMBeanReference(ObjectName objectName, HashMap mbeanReferenceMap)
	{
		synchronized(m_referencedMBeanObjectNameToRelationIds)
		{
			m_referencedMBeanObjectNameToRelationIds.put(objectName, mbeanReferenceMap);
		}
	}

	// remove a relation from the RelationService
	public void removeRelation(String relationId) throws IllegalArgumentException, RelationServiceNotRegisteredException,
		                                   RelationNotFoundException
	{
		isActive();
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("Removing a Relation from the RelationService.");
		Object result = getRelationObject(relationId);
		if(result instanceof ObjectName)
		{
			// add the objectName to List
			List obsoleteReferences = new ArrayList();
			obsoleteReferences.add((ObjectName)result);
			// update Listener with the list of ObjectNames to be deregistered
			updateUnregistrationListener(null, obsoleteReferences);
		}
		// notify all listeners a relation has been removed
		sendRelationRemovalNotification(relationId, null);

		List referencedMBeanList = new ArrayList();
		List nonReferencedObjectNameList = getNonReferencedMBeans(relationId);
		// remove ObjectNames from global Map
		removeNonReferencedMBeans(nonReferencedObjectNameList);
		// remove relationId for our nonReferencedMBeans
		removeRelationId(relationId);
		// we have an MBean remove it from our Map
		if(result instanceof ObjectName)removeRelationMBean((ObjectName)result);
		// get the corresponding relationTypeName for the relationId
		String relationTypeName = getRelationTypeNameFromMap(relationId);
		// remove relationId from our Map
		removeRelationIdToRelationTypeName(relationId);
		// get all the corresponding relationIds
		List relationIdsList = getRelationIds(relationTypeName);

		// we have relationIds for the removed relationTypeName remove them from our Map
		if(relationIdsList != null)
		{
			relationIdsList.remove(relationId);
			if(relationIdsList.isEmpty())
			{
				// now we can remove the relationTypeName
				removeRelationTypeName(relationTypeName);
			}
		}
	}

	private void removeRelationMBean(ObjectName objectName)
	{
		synchronized(m_relationMBeanObjectNameToRelationId)
		{
			m_relationMBeanObjectNameToRelationId.remove(objectName);
		}
	}

	private void removeNonReferencedMBeans(List nonReferencedMBeansList)
	{
		synchronized(m_referencedMBeanObjectNameToRelationIds)
		{
			for(Iterator i = nonReferencedMBeansList.iterator(); i.hasNext();)
			{
				ObjectName currentObjectName = (ObjectName)i.next();
				m_referencedMBeanObjectNameToRelationIds.remove(currentObjectName);
			}
		}
	}

	private List getNonReferencedMBeans(String relationId)
	{
		List referencedMBeanList = new ArrayList();
		List nonReferencedObjectNameList = new ArrayList();
		synchronized(m_referencedMBeanObjectNameToRelationIds)
		{
			for(Iterator i = (m_referencedMBeanObjectNameToRelationIds.keySet()).iterator(); i.hasNext();)
			{
				ObjectName currentObjectName = (ObjectName)i.next();
				HashMap relationIdMap = (HashMap)(m_referencedMBeanObjectNameToRelationIds.get(currentObjectName));
				if(relationIdMap.containsKey(relationId))
				{
					relationIdMap.remove(relationId);
					referencedMBeanList.add(currentObjectName);
				}
				if(relationIdMap.isEmpty()) nonReferencedObjectNameList.add(currentObjectName);
			}
		}
		return nonReferencedObjectNameList;
	}

	private void removeRelationTypeName(String relationTypeName)
	{
		synchronized(m_relationTypeNameToRelationIds)
		{
			m_relationTypeNameToRelationIds.remove(relationTypeName);
		}
	}

	private String getRelationTypeNameFromMap(String relationId)
	{
		synchronized(m_relationIdToRelationTypeName)
		{
		    return ((String)m_relationIdToRelationTypeName.get(relationId));
		}
	}

	private void removeRelationIdToRelationTypeName(String relationId)
	{
		synchronized(m_relationIdToRelationTypeName)
		{
		    m_relationIdToRelationTypeName.remove(relationId);
		}
	}

	private String getMBeanObjectName(ObjectName objectName)
	{
		synchronized(m_relationMBeanObjectNameToRelationId)
		{
			return ((String)m_relationMBeanObjectNameToRelationId.get(objectName));
		}
	}

	private void addMBeanObjectNameToRelationId(ObjectName relationObjectName, String relationId)
	{
		synchronized(m_relationMBeanObjectNameToRelationId)
		{
		    m_relationMBeanObjectNameToRelationId.put(relationObjectName, relationId);
		}
	}

	private void removeRelationId(String relationId)
	{
		synchronized(m_relationIdToRelationObject)
		{
		    m_relationIdToRelationObject.remove(relationId);
		}
	}

	private Map getAllMBeanObjectNames(ObjectName objectName)
	{
		synchronized(m_relationMBeanObjectNameToRelationId)
		{
			return ((HashMap)m_relationMBeanObjectNameToRelationId.get(objectName));
		}
	}

	private void updateUnregistrationListener(List newReferenceList, List obsoleteReferences) throws RelationServiceNotRegisteredException
	{
		Logger logger = getLogger();
		if (newReferenceList != null && obsoleteReferences != null)
		{
			// nothing to do
	        if (newReferenceList.isEmpty() && obsoleteReferences.isEmpty())
			{
				return;
			}
		}
		isActive();
		if (newReferenceList != null || obsoleteReferences != null)
		{
			boolean isNewListener = false;
			if(m_notificationFilter == null)
			{
				m_notificationFilter = new MBeanServerNotificationFilter();
				isNewListener = true;
			}
			synchronized(m_notificationFilter)
			{
				// we have new references - update (Enable in NotificationFilter)
				if(newReferenceList != null)
				{
					updateNewReferences(newReferenceList);
				}
				// we have obsolete references - update (disable from notificationFilter)
				if(obsoleteReferences != null)
				{
					updateObsoleteReferences(obsoleteReferences);
				}

				ObjectName mbeanServerDelegateName = null;
				try	{mbeanServerDelegateName = new ObjectName("JMImplementation:type=MBeanServerDelegate");}
				catch (MalformedObjectNameException ignored) {}

				if(isNewListener)
				{
					try
					{
						m_server.addNotificationListener(mbeanServerDelegateName, this, m_notificationFilter, null);
					}
					catch (InstanceNotFoundException ex)

					{
						throw new RelationServiceNotRegisteredException(ex.getMessage());
					}
				}
			}
		}
	}

	private void updateObsoleteReferences(List obsoleteReferences)
	{
		for(Iterator i = obsoleteReferences.iterator(); i.hasNext();)
		{
			ObjectName name = (ObjectName)i.next();
			m_notificationFilter.disableObjectName(name);
		}
	}

	private void updateNewReferences(List newReferencesList)
	{
		for(Iterator i = newReferencesList.iterator(); i.hasNext();)
		{
			ObjectName name = (ObjectName) i.next();
			m_notificationFilter.enableObjectName(name);
		}
	}

	// if null we have an instance of the internalRelation otherwise we have an MBean and can use the proxy
	private Relation getRelation(String relationId) throws RelationNotFoundException
	{
		if(relationId == null)throw new IllegalArgumentException("Null relation id passed into getRelation.");
		Relation relation = null;
		// if this is null we have a Relation not an ObjectName so we can return the Relation
		if(isRelationMBean(relationId) == null)
		{
			synchronized(m_relationIdToRelationObject)
		    {
				relation = (Relation)m_relationIdToRelationObject.get(relationId);
				return relation;
			}
		}
		if(!(m_proxy == null)) return m_proxy;
		// we found no relation and the m_proxy is null we have an error condition
		throw new RelationNotFoundException("Relation not found with ID: " + relationId);
	}

	// will return an ObjectName or a Relation object
	private Object getRelationObject(String relationId)throws IllegalArgumentException, RelationNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		Object relationObject = null;
		synchronized(m_relationIdToRelationObject)
		{
		    relationObject = m_relationIdToRelationObject.get(relationId);
			if(relationObject == null)
			{
				// if this happens is caught by method looking for a null.
				throw new RelationNotFoundException("Null Relation");
			}
			// we return either an ObjectName or an InternalRelation
			return relationObject;
		}
	}

	public void purgeRelations() throws RelationServiceNotRegisteredException
	{
		isActive();
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("purging relations");

		ArrayList localDeregisteredNotificationList = null;
		synchronized(m_deregisteredNotificationList)
		{
			// cone the list of notifications to be able to recieve notification while processing current ones
			localDeregisteredNotificationList = ((ArrayList)((ArrayList)m_deregisteredNotificationList).clone());
			// now reset
			m_deregisteredNotificationList = new ArrayList();
		}

		List obsoleteReferenceList = new ArrayList();
		Map localMBeanToRelationId = new HashMap();
		synchronized(m_referencedMBeanObjectNameToRelationIds)
		{
			for(Iterator i = localDeregisteredNotificationList.iterator(); i.hasNext();)
			{
				MBeanServerNotification serverNotification = (MBeanServerNotification)i.next();
				ObjectName deregisteredMBeanName = serverNotification.getMBeanName();
				obsoleteReferenceList.add(deregisteredMBeanName);

				HashMap relationIdMap = (HashMap) m_referencedMBeanObjectNameToRelationIds.get(deregisteredMBeanName);
				localMBeanToRelationId.put(deregisteredMBeanName, relationIdMap);
				m_referencedMBeanObjectNameToRelationIds.remove(deregisteredMBeanName);
			}
		}

		// update listener filter to avoid recieving notifications for same MBeans
		updateUnregistrationListener(null, obsoleteReferenceList);
		for(Iterator i = localDeregisteredNotificationList.iterator(); i.hasNext();)
		{
			MBeanServerNotification currentNotification = (MBeanServerNotification)i.next();
			ObjectName unregisteredMBeanObjectName = currentNotification.getMBeanName();
			HashMap localRelationIdMap = (HashMap)(localMBeanToRelationId.get(unregisteredMBeanObjectName));

			Set localRelationIdSet = localRelationIdMap.keySet();
			// handles the unregistration of mbeans
			unregisterReferences(localRelationIdSet, localRelationIdMap, unregisteredMBeanObjectName);
		}
	}

	private void unregisterReferences(Set relationIdSet, Map relationIdMap, ObjectName objectName)
			throws RelationServiceNotRegisteredException
	{
		for(Iterator iter = relationIdSet.iterator(); iter.hasNext();)
		{
			String currentRelationId = (String)iter.next();
			ArrayList localRoleNamesList = (ArrayList)(relationIdMap.get(currentRelationId));
			try
			{
				handleReferenceUnregistration(currentRelationId, objectName, localRoleNamesList);
			}
			catch (RelationTypeNotFoundException ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
			catch (RelationNotFoundException ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
			catch(RoleNotFoundException ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
		}
	}

	private void handleReferenceUnregistration(String relationId, ObjectName unregisteredObjectName, List roleNames)
			throws IllegalArgumentException, RelationServiceNotRegisteredException, RelationNotFoundException,
			RoleNotFoundException, RelationTypeNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Null relationId");
		if(unregisteredObjectName == null) throw new IllegalArgumentException("Null ObjectName");
		if(roleNames == null) throw new IllegalArgumentException("Null roleName list");

		isActive();
		String relationTypeName = getRelationTypeName(relationId);
		Object relationObject = getRelationObject(relationId);
		boolean canDeleteRelation = false;
		for(Iterator i = roleNames.iterator(); i.hasNext();)
		{
			String currentRoleName = (String)i.next();
			int currentRoleCardinality = (getRoleCardinality(relationId, currentRoleName)).intValue();
			int newRoleCardinality = currentRoleCardinality - 1;
			RoleInfo currentRoleInfo = null;
			try
			{
				currentRoleInfo = getRoleInfo(relationTypeName, currentRoleName);
			}
			catch (RelationTypeNotFoundException ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
			catch (RoleInfoNotFoundException ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
			// check that the role cardinality is maintained by the removal
			if(!(currentRoleInfo.checkMinDegree(newRoleCardinality)))
			{
				canDeleteRelation = true;
			}
		}
		// roleMinValue been checked everything ok, can now remove the relation
		if(canDeleteRelation)
		{
			removeRelation(relationId);
		}
		else
		{
			for(Iterator i = roleNames.iterator(); i.hasNext();)
			{
				String currentRoleName = (String)i.next();
				try
				{
					Relation relation = getRelation(relationId);
					relation.handleMBeanUnregistration(unregisteredObjectName, currentRoleName);
				}
				catch(InvalidRoleValueException ex){ throw new RuntimeOperationsException(ex.getMessage());}
			}
		}
	}

	private void removeObjectName(ObjectName objectName)
	{
		synchronized(m_referencedMBeanObjectNameToRelationIds)
		{
			m_referencedMBeanObjectNameToRelationIds.remove(objectName);
		}
	}

	public Map findReferencingRelations(ObjectName mbeanObjectName, String relationTypeName, String roleName) throws
		                                   IllegalArgumentException
	{
		if(mbeanObjectName == null) throw new IllegalArgumentException("Cannot find references for a null ObjectName");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("finding referencing relations for MBean with ObjectName: "
		           + mbeanObjectName.getCanonicalName() + " and relationTypeName: " + relationTypeName + " and roleName: " + roleName);
		HashMap result = new HashMap();
		HashMap relationIdMap = (HashMap)getReferencedMBeansFromMap(mbeanObjectName);
		if(relationIdMap != null)
		{
		    Set allRelationIds = relationIdMap.keySet();
			List relationIdList = null;
			if(relationTypeName == null) relationIdList = new ArrayList(allRelationIds);
			else relationIdList = findReferencesFromIds(allRelationIds, relationTypeName);

			for(Iterator i = relationIdList.iterator(); i.hasNext();)
			{
				String currentRelationId = (String)i.next();
				ArrayList currentRoleNameList = (ArrayList)relationIdMap.get(currentRelationId);
				if(roleName == null) result.put(currentRelationId, currentRoleNameList.clone());
				else if(currentRoleNameList.contains(roleName))
				{
					ArrayList roleNameList = new ArrayList();
					roleNameList.add(roleName);
					result.put(currentRelationId, roleNameList);
				}
			}
		}
		return result;
	}

	private ArrayList findReferencesFromIds(Set allRelationIds, String relationTypeName)
	{
		ArrayList relationIdList = new ArrayList();
		for(Iterator i = allRelationIds.iterator(); i.hasNext();)
		{
			String currentRelationId = (String)i.next();
			String currentRelationTypeName = getRelationTypeNameFromMap(currentRelationId);
			if(currentRelationTypeName.equals(relationTypeName))
			{
				relationIdList.add(currentRelationId);
			}
		}
		return relationIdList;
	}

	public Map findAssociatedMBeans(ObjectName mbeanObjectName, String relationTypeName, String roleName) throws
			                               IllegalArgumentException
	{
		if(mbeanObjectName == null)throw new IllegalArgumentException("mbean ObjectName cannot be null.");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("finding associated relations for MBean with ObjectName: "
		           + mbeanObjectName.getCanonicalName() + " and relationTypeName: " + relationTypeName + " and roleName: " + roleName);
		Map relationIdsToRoleNames = (HashMap)findReferencingRelations(mbeanObjectName, relationTypeName, roleName);
		Map result = new HashMap();
		for(Iterator i = (relationIdsToRoleNames.keySet()).iterator(); i.hasNext();)
		{
			String currentRelationId = (String)i.next();
			HashMap objectNamesToRoleMap = null;
			try
			{
				objectNamesToRoleMap = (HashMap)(getReferencedMBeans(currentRelationId));
			}
			catch (RelationNotFoundException ex)
			{
				logger.error("Relation with ID: " + currentRelationId + " not found.");
				throw new RuntimeOperationsException("Relation Not Found");
			}
			for(Iterator iter = (objectNamesToRoleMap.keySet()).iterator(); iter.hasNext();)
			{
				ObjectName objectName = (ObjectName)iter.next();
				if(!(objectName.equals(mbeanObjectName)))
				{
					ArrayList currentRelationIdList = new ArrayList();
					currentRelationIdList.add(currentRelationId);
					result.put(objectName, currentRelationIdList);
				}
			}
		}
		return result;
	}

	public List findRelationsOfType(String relationTypeName) throws IllegalArgumentException, RelationTypeNotFoundException
	{
		if(relationTypeName == null) throw new IllegalArgumentException("relation type name cannot be null.");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("finding relations matching relationTypeName: " + relationTypeName);
		RelationType relationType = getRelationType(relationTypeName);
		List relationIdList = getRelationIds(relationTypeName);
		List result = new ArrayList();
		if(relationIdList != null)
		{
			// no need to clone read only method
			result = relationIdList;
		}
		return result;
	}

	public List getRole(String relationId, String roleName) throws IllegalArgumentException, RelationServiceNotRegisteredException,
			RelationNotFoundException, RoleNotFoundException
	{
		isActive();
		if(relationId == null) throw new IllegalArgumentException("RelationId cannot have a null value.");
		if(roleName == null) throw new IllegalArgumentException("Role Name cannot have a null value.");
		Relation relationObject = getRelation(relationId);
		return relationObject.getRole(roleName);
	}

	public RoleResult getRoles(String relationId, String[] roleNames) throws IllegalArgumentException, RelationNotFoundException, RelationServiceNotRegisteredException
	{
		if(relationId == null) throw new IllegalArgumentException("Illegal Argument relationId is null.");
		if(roleNames == null) throw new IllegalArgumentException("Array of Roles Names should not be null");
		isActive();
		Relation relation = getRelation(relationId);
		return relation.getRoles(roleNames);
	}

	public RoleResult getAllRoles(String relationId) throws IllegalArgumentException, RelationNotFoundException,
		                                   RelationServiceNotRegisteredException
	{
		if(relationId == null) throw new IllegalArgumentException("RelationId cannot be null");
		//Object relationObject = getRelationObject(relationId);
		Relation relation = getRelation(relationId);
		return relation.getAllRoles();
	}

	public Integer getRoleCardinality(String relationId, String roleName)throws IllegalArgumentException,
				                           RelationNotFoundException, RoleNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Relation Id is null.");
		if(roleName == null) throw new IllegalArgumentException("Role Name is null.");
		Object relationObject = getRelationObject(relationId);
		if(relationObject instanceof InternalRelation) return (((InternalRelation)relationObject).getRoleCardinality(roleName));
		else return m_proxy.getRoleCardinality(roleName);
	}

	public void setRole(String relationId, Role role) throws IllegalArgumentException, RelationServiceNotRegisteredException,
		                                   RelationNotFoundException, RoleNotFoundException, InvalidRoleValueException, RelationTypeNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Illegal Null Relation Id.");
		if(role == null) throw new IllegalArgumentException("Illegal Null Role.");
		isActive();
		Relation relation = getRelation(relationId);
		relation.setRole(role);
	}

	public RoleResult setRoles(String relationId, RoleList roleList) throws RelationServiceNotRegisteredException,
		                                   IllegalArgumentException, RelationNotFoundException
	{
		Logger logger = getLogger();
		if(relationId == null) throw new IllegalArgumentException("Relation Id is null");
		if(roleList == null) throw new IllegalArgumentException("RoleList is null");
		isActive();
		Object relationObject = getRelationObject(relationId);
		RoleResult roleResult = null;
		Relation relation = getRelation(relationId);
		try
		{
		    return relation.setRoles(roleList);
		}
		catch(RelationTypeNotFoundException ex)
		{
			throw new RuntimeOperationsException("Unable to find a RelationTypeName for relation ID: " + relationId);
		}
	}

	public Map getReferencedMBeans(String relationId) throws IllegalArgumentException, RelationNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("getting MBeans referenced for ID: " + relationId);
		Relation relation = getRelation(relationId);
		return relation.getReferencedMBeans();
	}

	public String getRelationTypeName(String relationId) throws IllegalArgumentException, RelationNotFoundException
	{
		if(relationId == null) throw new IllegalArgumentException("Null Relation Id");
		Logger logger = getLogger();
		if(logger.isEnabledFor(Logger.INFO))logger.info("getting the relationType for ID: " + relationId);
		Relation relation = getRelation(relationId);
		return relation.getRelationTypeName();
	}

	public void handleNotification(Notification notification, Object handback)
	{
		if(notification == null) throw new IllegalArgumentException("Null Notification");
		if(notification instanceof MBeanServerNotification)
		{
			String notificationType = notification.getType();
			if(notificationType.equals(MBeanServerNotification.UNREGISTRATION_NOTIFICATION))
			{
				ObjectName mbeanName = ((MBeanServerNotification)notification).getMBeanName();
				// handle unregistration by purging relations
				handleUnregistration(notification, mbeanName);
				// remove mbeans now
				handleMBeanRemoval(mbeanName);
			}
		}
	}

	private void handleMBeanRemoval(ObjectName mbeanName)
	{
		String relationId = getMBeanObjectName(mbeanName);
		if(relationId != null)
		{
			try
			{
				// we have a relationId now removeRelation
				removeRelation(relationId);
			}
			catch (Exception ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
		}
	}

	private void handleUnregistration(Notification notification, ObjectName objectName)
	{
		boolean isReferenced = false;
		synchronized(m_referencedMBeanObjectNameToRelationIds)
		{
			// check the mbeanObjectName can be found
			if(m_referencedMBeanObjectNameToRelationIds.containsKey(objectName))
			{
				// it does so add it to our list of deregistrations
				synchronized(m_deregisteredNotificationList)
				{
					m_deregisteredNotificationList.add(notification);
				}
				isReferenced = true;
			}
			if(isReferenced && m_purgeFlag)
			{
				try
				{
					// purge the RelationService
					purgeRelations();
				}
				catch (Exception ex)
				{
					throw new RuntimeOperationsException(ex.getMessage());
				}
			}
		}
	}

	/**
	 * Implement to return more info regarding Notification Types
	 */
	public MBeanNotificationInfo[] getNotificationInfo()
	{
		MBeanNotificationInfo[] notificationInfo = new MBeanNotificationInfo[1];
		String[] notificationTypes = new String[6];
		notificationTypes[0] = RelationNotification.RELATION_BASIC_CREATION;
		notificationTypes[1] = RelationNotification.RELATION_MBEAN_CREATION;
		notificationTypes[2] = RelationNotification.RELATION_BASIC_UPDATE;
		notificationTypes[3] = RelationNotification.RELATION_MBEAN_UPDATE;
		notificationTypes[4] = RelationNotification.RELATION_BASIC_REMOVAL;
		notificationTypes[5] = RelationNotification.RELATION_MBEAN_REMOVAL;

		String notificationDescription = "Sent when a relation is created, updated or deleted.";
		notificationInfo[0] = new MBeanNotificationInfo(notificationTypes, "RelationNotification", notificationDescription);
		return notificationInfo;
	}

	private void addDeregisteredMbean(Notification notification)
	{
		synchronized(m_deregisteredNotificationList)
		{
			m_deregisteredNotificationList.add(notification);
		}
	}

	private boolean isMBeanReferenced(ObjectName mbeanObjectName)
	{
		synchronized(m_relationMBeanObjectNameToRelationId)
		{
			return (m_relationMBeanObjectNameToRelationId.containsKey(mbeanObjectName));
		}
	}

	public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception
	{
		m_server = server;
		m_relationServiceObjectName = name == null ? new ObjectName(m_server.getDefaultDomain(), "service", "Relation") : name;
		return m_relationServiceObjectName;
	}

	public void postRegister(Boolean registrationDone)
	{
		Logger logger = getLogger();
		boolean done = registrationDone.booleanValue();
		if (!done)
		{
			m_server = null;
			logger.warn("Relation service was NOT registered");
		}
		else
		{
			if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Relation service postRegistered");}
		}
	}

	public void preDeregister() throws Exception
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) {logger.debug("Relation service preDeregistered");}
	}

	public void postDeregister()
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) {logger.debug("Relation service postDeregistered");}
	}

	static void throwRoleProblemException(int problemType, String roleName) throws IllegalArgumentException,
		          RoleNotFoundException, InvalidRoleValueException
	{
		if(roleName == null) throw new IllegalArgumentException("Null Role Name");
		switch(problemType)
		{
			case RoleStatus.NO_ROLE_WITH_NAME:
				 throw new RoleNotFoundException("RoleName: " + roleName + " does not exist in the relation");
			case RoleStatus.ROLE_NOT_READABLE:
				 throw new RoleNotFoundException("RoleName: " + roleName + " is not readable");
			case RoleStatus.ROLE_NOT_WRITABLE:
				 throw new RoleNotFoundException("RoleName: " + roleName + " is not writable");
			case RoleStatus.LESS_THAN_MIN_ROLE_DEGREE:
				 throw new InvalidRoleValueException("RoleName: " + roleName +
					   " has references less than the expected minimum.");
			case RoleStatus.MORE_THAN_MAX_ROLE_DEGREE:
				 throw new InvalidRoleValueException("RoleName: " + roleName +
					   " has references more than the expected maximum.");
			case RoleStatus.REF_MBEAN_OF_INCORRECT_CLASS:
				 throw new InvalidRoleValueException("RoleName: " + roleName +
					   " has a MBean reference to a MBean not of the expected class of references for that role.");
			case RoleStatus.REF_MBEAN_NOT_REGISTERED:
				 throw new InvalidRoleValueException("RoleName: " + roleName +
					   "  has a reference to null MBean or to a MBean not registered.");
		}
	}

	private Logger getLogger()
	{
		return Log.getLogger("jmx.mbean." + getClass().getName());
	}

	// inner class to represent internal relations defined in the relationService
	final class InternalRelation extends RelationSupport
	{
		InternalRelation(String relationId, ObjectName relationServiceObjectName, String relationTypeName,
		   RoleList roleList) throws InvalidRoleValueException, IllegalArgumentException
		{
			super(relationId, relationServiceObjectName, relationTypeName, roleList);
		}

		// checks role can be read
		int getReadingProblemType(Role role, String roleName, String relationTypeName)
		{
			if(roleName == null) throw new IllegalArgumentException("Null RoleName");
			if(role == null) return(RoleStatus.NO_ROLE_WITH_NAME);
			else
			{
				try
				{
					return ((checkRoleReading(roleName, relationTypeName)).intValue());
				}
				catch(RelationTypeNotFoundException ex){throw new RuntimeOperationsException(ex.getMessage());}
			}
		}

		// overridden from RelationSupport
		int getRoleWritingValue(Role role, String relationTypeName, Boolean toBeInitialized) throws RelationTypeNotFoundException
		{
			try
			{   // check role is writable
				return (checkRoleWriting(role, relationTypeName, toBeInitialized)).intValue();
			}
			catch(RelationTypeNotFoundException ex)
			{
				throw new RuntimeOperationsException(ex.getMessage());
			}
		}

		// sends role update notification
		void sendUpdateRoleNotification(String relationId, Role role, List oldRoleValue) throws RelationServiceNotRegisteredException,
			 RelationNotFoundException
		{
			if(relationId == null) throw new IllegalArgumentException("Null RelationId passed into sendUpdateRoleNotification");
		    if(role == null)throw new IllegalArgumentException("Null role passed into sendUpdateRoleNotification");
		    if(oldRoleValue == null)throw new IllegalArgumentException("Null list of role Values passed into sendUpdateRoleNotification");
			sendRoleUpdateNotification(relationId, role, oldRoleValue);
		}

		void updateRelationServiceMap(String relationId, Role role, List oldRoleValue)throws IllegalArgumentException,
			 RelationServiceNotRegisteredException, RelationNotFoundException
		{
			if(relationId == null) throw new IllegalArgumentException("Null RelationId passed into updateRelationServiceMap");
		    if(role == null)throw new IllegalArgumentException("Null role passed into updateRelationServiceMap");
		    if(oldRoleValue == null)throw new IllegalArgumentException("Null list of role Values passed into updateRelationServiceMap");
			updateRoleMap(relationId, role, oldRoleValue);
		}
	}
}