/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */
package mx4j.adaptor.http;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.JMException;
import javax.management.loading.DefaultLoaderRepository;
import java.io.IOException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import javax.xml.parsers.DocumentBuilder;
import java.lang.reflect.Constructor;

/**
 * ConstructorsCommandProcessor, processes a request to query the available
 * constructors for a classname
 *
 * @author <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version $Revision: 1.4 $
 */
class ConstructorsCommandProcessor implements HttpCommandProcessor
{
	private MBeanServer server;

	private DocumentBuilder builder;

	public ConstructorsCommandProcessor(MBeanServer server, DocumentBuilder builder)
	{
		this.server = server;
		this.builder = builder;
	}

	public Document executeRequest(HttpInputStream in) throws IOException, JMException
	{
		Document document = builder.newDocument();

		String classname = in.getVariable("classname");
		if (classname == null || classname.trim().length() == 0)
		{
			return createException(document, "classname parameter required");
		}
		else
		{
			// look class in default classloader
			Class targetClass = null;
			try
			{
				targetClass = DefaultLoaderRepository.loadClass(classname);
				if (targetClass == null)
				{
					targetClass = ClassLoader.getSystemClassLoader().loadClass(classname);
				}
				if (targetClass == null)
				{
					targetClass = getClass().getClassLoader().loadClass(classname);
				}
			}
			catch (ClassNotFoundException e)
			{
				return createException(document, "class " + classname + " not found");
			}

			Element root = document.createElement("Class");
			root.setAttribute("classname", classname);
			document.appendChild(root);
			Constructor[] constructors = targetClass.getConstructors();
			for (int i=0;i<constructors.length;i++)
			{
				Element constructor = document.createElement("Constructor");
				constructor.setAttribute("name", constructors[i].getName());
				addParameters(constructor, document, constructors[i].getParameterTypes());
				root.appendChild(constructor);
			}
		}
		return document;
	}

	protected void addParameters(Element node, Document document, Class[] parameters)
	{
		for (int j=0;j<parameters.length;j++)
		{
			Element parameter = document.createElement("Parameter");
			parameter.setAttribute("type", parameters[j].getName());
			// add id since order is relevant
			parameter.setAttribute("id", ""+j);
			node.appendChild(parameter);
		 }
	}

	private Document createException(Document document, String message)
	{
		Element exceptionElement = document.createElement("Exception");
		document.appendChild(exceptionElement);
		exceptionElement.setAttribute("errorMsg", message);
		return document;
	}

}
