/*
 *  Copyright (C) MX4J.
 *  All rights reserved.
 *
 *  This software is distributed under the terms of the MX4J License version 1.0.
 *  See the terms of the MX4J License in the documentation provided with this software.
 */
package mx4j.adaptor.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.InterruptedIOException;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.StringTokenizer;
import javax.management.JMException;
import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import mx4j.log.Log;
import mx4j.log.Logger;
import mx4j.adaptor.AdaptorServerSocketFactory;
import mx4j.adaptor.PlainAdaptorServerSocketFactory;
import org.w3c.dom.Document;

/**
 * HttpAdaptor sets the basic adaptor listening for HTTP requests
 *
 * @author    <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version   $Revision: 1.3 $
 */
public class HttpAdaptor implements HttpAdaptorMBean, MBeanRegistration
{
	/** Port to listen for connections */
	private int port = 8080;

	/** Host where to set the server socket */
	private String host = "localhost";

	/** Target server */
	private MBeanServer server;

	/** Server socket */
	private ServerSocket serverSocket;

	/** Indicates whether the server is running */
	private boolean alive;

	/** Internal logger */
	private Logger log;

	/** Map of commands indexed by the request path */
	private Map commands = new HashMap();

	/** Target processor */
	private ProcessorMBean processor = null;

	/** Target processor name */
	private ObjectName processorName = null;

	/** Default processor */
	private ProcessorMBean defaultProcessor = new DefaultProcessor();

	private String authenticationMethod = "none";

	// Should be dependant on the server?
	private String realm="MX4J";

	private Map authorizations = new HashMap();

	private AdaptorServerSocketFactory socketFactory = null;

	private String factoryName;

	/**
	 * Sets the value of the server's port
	 *
	 * @param port  the new port's value
	 */
	public void setPort(int port)
	{
		if (alive)
		{
			throw new IllegalArgumentException("Not possible to change port with the server running");
		}
		this.port = port;
	}


	/**
	 * Sets the host name where the server will be listening
	 *
	 * @param host  Server's host
	 */
	public void setHost(String host)
	{
		if (alive)
		{
			throw new IllegalArgumentException("Not possible to change port with the server running");
		}
		this.host = host;
	}

	/**
	 * Authentication Method
	 *
	 * @return authentication method
	 */
	public String getAuthenticationMethod()
	{
		return authenticationMethod;
	}

	/**
	 * Sets the Authentication Method.
	 *
	 * @param method none/basic/digest
	 */
	public void setAuthenticationMethod(String method) {
		if (alive)
		{
			throw new IllegalArgumentException("Not possible to change authentication method with the server running");
		}
		if (method == null || !(method.equals("none") || method.equals("basic") || method.equals("digest")))
		{
			throw new IllegalArgumentException("Only accept methods none/basic/digest");
		}
		this.authenticationMethod = method;
	}


	/**
	 * Sets the object which will post process the XML results. The last value set
	 * between the setPostProcessor and setPostProcessorName will be the valid one
	 *
	 * @param processor  a Post processor object
	 */
	public void setProcessor(ProcessorMBean processor)
	{
		this.processor = processor;
		this.processorName = null;
	}

	/**
	 * Sets the object which create the server sockets
	 *
	 * @param factory the socket factory
	 */
	public void setSocketFactory(AdaptorServerSocketFactory factory)
	{
		this.factoryName = null;
		this.socketFactory = factory;
	}

	public void setSocketFactoryName(String factoryName)
	{
		this.socketFactory = null;
		this.factoryName = factoryName;
	}

	/**
	 * Sets the object name which will post process the XML result. The last value
	 * set between the setPostProcessor and setPostProcessorName will be the valid
	 * one. The MBean will be verified to be of instance HttpPostProcessor
	 *
	 * @param processorName  The new processorName value
	 */
	public void setProcessorName(ObjectName processorName)
	{
		this.processor = null;
		this.processorName = processorName;
	}

	public ProcessorMBean getProcessor()
	{
		return this.processor;
	}

	public ObjectName getProcessorName()
	{
		return this.processorName;
	}


	/**
	 * Returns the port where the server is running on. Default is 8080
	 *
	 * @return   HTTPServer's port
	 */
	public int getPort()
	{
		return port;
	}


	/**
	 * Return the host name the server will be listening to. If null the server
	 * listen at the localhost
	 *
	 * @return   the current hostname
	 */
	public String getHost()
	{
		return host;
	}


	/**
	 * Indicates whether the server's running
	 *
	 * @return        The active value
	 */
	public boolean isActive()
	{
		return alive;
	}


	/**
	 * Gathers some basic data
	 */
	public ObjectName preRegister(MBeanServer server, ObjectName name)
		throws java.lang.Exception
	{
		this.server = server;
		log = Log.getLogger(name.toString());
		buildCommands();
		return name;
	}


	public void postRegister(Boolean registrationDone) { }


	public void preDeregister()
		throws java.lang.Exception
	{
		// stop the server
		stop();
	}


	public void postDeregister() { }


	private ServerSocket createServerSocket() throws IOException
	{
		if (socketFactory == null)
		{
			if (factoryName == null)
			{
				socketFactory = new PlainAdaptorServerSocketFactory();
				return socketFactory.createServerSocket(port, 50, host);
			}
			else
			{
				try
				{
					return (ServerSocket)server.invoke(new ObjectName(factoryName), "createServerSocket", new Object[] {new Integer(port), new Integer(50), host}, new String[] {"int", "int", "java.lang.String"});
				}
				catch (Exception x)
				{
					log.error("Exception invoking AdaptorServerSocketFactory via MBeanServer", x);
				}
			}
		}
		else
		{
			return socketFactory.createServerSocket(port, 50, host);
		}

		return null;
	}


	/**
	 * Starts the server
	 */
	public void start()
		throws IOException
	{
		if (server != null)
		{
			serverSocket = createServerSocket();

			if (serverSocket == null)
			{
				log.error("Server socket is null");
				return;
			}

			log.info("HttpAdaptor server listening on port " + port);
			alive = true;
			Thread serverThread = new Thread(
				new Runnable()
				{
					public void run()
					{
						log.info("HttpAdaptor starting");
						while (alive)
						{
							try
							{
								Socket client = null;
								client = serverSocket.accept();
								if (!alive)
								{
									break;
								}
								new HttpClient(client).start();
							}
							catch (InterruptedIOException e)
							{
								continue;
							}
							catch (IOException e)
							{
								continue;
							}
							catch (Exception e)
							{
								log.warn("Exception during request processing", e);
								continue;
							}
							catch (Error e)
							{
								log.error("Error during request processing", e);
								continue;
							}
						}
						try
						{
							serverSocket.close();
						}
						catch (IOException e)
						{
							log.warn("Exception closing the server", e);
						}
						serverSocket = null;
						alive = false;
						log.info("Server stopped");
					}
				});
			serverThread.start();
		}
		else
		{
			log.info("Start failed, no server target server has been set");
		}
	}


	/**
	 * Restarts the server. Useful when changing the Server parameters
	 *
	 * @deprecated as of RC 1
	 */
	public void restart()
		throws IOException
	{
		stop();
		start();
	}


	/**
	 * Stops the HTTP daemon
	 */
	public void stop()
	{
		try
		{
			if (alive)
			{
				alive = false;
				// force the close with a socket call
				new Socket(host, port);
				if (serverSocket != null)
				{
					serverSocket.close();
				}
			}
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}
	}


	public void addAuthorization(String username, String password)
	{
		if (username == null || password == null)
		{
			throw new IllegalArgumentException("username and passwords cannot be null");
		}
		authorizations.put(username, password);
	}

	private boolean isUsernameValid(String username, String password)
	{
		if (authorizations.containsKey(username))
		{
			return password.equals(authorizations.get(username));
		}
		return false;
	}


	/**
	 * Gets a processor for a given path
	 */
	protected HttpCommandProcessor getProcessor(String path)
	{
		return (HttpCommandProcessor)commands.get(path);
	}


	/**
	 * Build the commands
	 */
	protected void buildCommands()
	{
		try
		{
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder = factory.newDocumentBuilder();
			commands.put("server", new ServerCommandProcessor(server, builder));
			commands.put("serverbydomain", new ServerByDomainCommandProcessor(server, builder));
			commands.put("mbean", new MBeanCommandProcessor(server, builder));
			commands.put("setattribute", new SetAttributeCommandProcessor(server, builder));
			commands.put("delete", new DeleteMBeanCommandProcessor(server, builder));
			commands.put("invoke", new InvokeOperationCommandProcessor(server, builder));
			commands.put("create", new CreateMBeanCommandProcessor(server, builder));
			commands.put("relation", new RelationCommandProcessor(server, builder));
			commands.put("constructors", new ConstructorsCommandProcessor(server, builder));
			//commands.put("mlet", new MLetCommandProcessor(server, builder));
		}
		catch (ParserConfigurationException e)
		{
			log.error("Exception building the Document Factories", e);
		}
	}


	protected void postProcess(HttpOutputStream out, HttpInputStream in, Document document)
		throws IOException, JMException
	{
		boolean processed = false;
		// inefficient but handles modifications at runtime
		if (processorName != null)
		{
			if (server.isRegistered(processorName) &&
					server.isInstanceOf(processorName, "mx4j.adaptor.http.ProcessorMBean"))
			{
				server.invoke(processorName,
						"writeResponse",
						new Object[]{out, in, document},
						new String[]{"mx4j.adaptor.http.HttpOutputStream", "mx4j.adaptor.http.HttpInputStream", "org.w3c.dom.Document"});
				processed = true;
			}
		}
		if (!processed && processor != null)
		{
			processor.writeResponse(out, in, document);
			processed = true;
		}
		if (!processed)
		{
			defaultProcessor.writeResponse(out, in, document);
		}
	}


	protected void findUnknownElement(String path, HttpOutputStream out, HttpInputStream in)
		throws IOException, JMException
	{
		boolean processed = false;
		// inefficient but handles modifications at runtime
		if (processorName != null)
		{
			if (server.isRegistered(processorName) &&
					server.isInstanceOf(processorName, "mx4j.adaptor.http.ProcessorMBean"))
			{
				server.invoke(processorName,
						"notFoundElement",
						new Object[]{path, out, in},
						new String[]{"java.lang.String", "mx4j.adaptor.http.HttpOutputStream", "mx4j.adaptor.http.HttpInputStream"});
				processed = true;
			}
		}
		if (!processed && processor != null)
		{
			processor.notFoundElement(path, out, in);
			processed = true;
		}
		if (!processed)
		{
			defaultProcessor.notFoundElement(path, out, in);
		}
	}


	protected String preProcess(String path)
		throws IOException, JMException
	{
		boolean processed = false;
		// inefficient but handles modifications at runtime
		if (processorName != null)
		{
			log.trace("Preprocess using " + processorName);
			if (server.isRegistered(processorName) &&
					server.isInstanceOf(processorName, "mx4j.adaptor.http.ProcessorMBean"))
			{
				log.trace("Preprocessing");
				path = (String)server.invoke(processorName,
						"preProcess",
						new Object[]{path},
						new String[]{"java.lang.String"});
				processed = true;
			}
			log.trace(processorName + " not found");
		}
		if (!processed && processor != null)
		{
			path = processor.preProcess(path);
			processed = true;
		}
		if (!processed)
		{
			path = defaultProcessor.preProcess(path);
		}
		return path;
	}


	protected void postProcess(HttpOutputStream out, HttpInputStream in, Exception e)
		throws IOException, JMException
	{
		boolean processed = false;
		// inefficient but handles modifications at runtime
		if (processorName != null)
		{
			if (server.isRegistered(processorName) &&
					server.isInstanceOf(processorName, "mx4j.adaptor.http.ProcessorMBean"))
			{
				server.invoke(processorName,
						"writeError",
						new Object[]{out, in, e},
						new String[]{"mx4j.adaptor.http.HttpOutputStream", "mx4j.adaptor.http.HttpInputStream", "java.lang.Exception"});
				processed = true;
			}
		}
		if (!processed && processor != null)
		{
			processor.writeError(out, in, e);
			processed = true;
		}
		if (!processed)
		{
			defaultProcessor.writeError(out, in, e);
		}
	}


	private class HttpClient extends Thread
	{
		private Socket client;


		HttpClient(Socket client)
		{
			this.client = client;
		}

		public boolean isValid(String authorizationString) {
			if (authenticationMethod.startsWith("basic"))
			{
				authorizationString = authorizationString.substring(5,authorizationString.length());
				String decodeString = Base64.decodeToString(authorizationString);
				if (decodeString.indexOf(":")>0)
				{
					try
					{
						StringTokenizer tokens = new StringTokenizer(decodeString, ":");
						String username = tokens.nextToken();
						String password = tokens.nextToken();
						return isUsernameValid(username, password);
					} catch (Exception e) {
						return false;
					}
				}
			}
			return false;
		}

		private boolean handleAuthentication(HttpInputStream in, HttpOutputStream out) throws IOException {
			if (authenticationMethod.equals("basic"))
			{
				Iterator h = in.getHeaders().keySet().iterator();
				String result = in.getHeader("authorization");
				if (result != null)
				{
					if (isValid(result))
					{
						return true;
					}
					throw new HttpException(HttpConstants.STATUS_FORBIDDEN, "Authentication failed");
				}

				out.setCode(HttpConstants.STATUS_AUTHENTICATE);
				out.setHeader("WWW-Authenticate", "Basic realm=\"" + realm + "\"");
				out.sendHeaders();
				out.flush();
				return false;
			}
			if (authenticationMethod.equals("digest"))
			{
			}
			return true;
		}


		public void run()
		{
			HttpInputStream httpIn = null;
			HttpOutputStream httpOut = null;
			try
			{
				// get input streams
				InputStream in = client.getInputStream();
				httpIn = new HttpInputStream(in);
				httpIn.readRequest();

				// Find a suitable command processor
				String path = httpIn.getPath();
				String queryString = httpIn.getQueryString();
				log.info("Request " + path + ((queryString == null) ? "" : ("?" + queryString)));
				String postPath = preProcess(path);
				if (!postPath.equals(path))
				{
					log.info("Processor replaced path " + path + " with the path " + postPath);
					path = postPath;
				}
				OutputStream out = client.getOutputStream();
				httpOut = new HttpOutputStream(out, httpIn);
				if (!handleAuthentication(httpIn, httpOut)) {
					return;
				}
				HttpCommandProcessor processor = getProcessor(path.substring(1, path.length()));
				if (processor == null)
				{
					log.info("No suitable command processor found, requesting from processor path " + path);
					findUnknownElement(path, httpOut, httpIn);
				}
				else
				{
					Document document = processor.executeRequest(httpIn);
					postProcess(httpOut, httpIn, document);
				}
			}
			catch (Exception ex)
			{
				log.warn("Exception during http request", ex);
				if (httpOut != null)
				{
					try
					{
						postProcess(httpOut, httpIn, ex);
					}
					catch (IOException e)
					{
						log.warn("IOException during http request", e);
					}
					catch (JMException e)
					{
						log.warn("JMException during http request", e);
					}
				}
			}
			catch (Error ex)
			{
				log.error("Error during http request ", ex);
			}
			finally
			{
				try
				{
					if (httpOut != null)
					{
						httpOut.flush();
					}
					// always close the socket
					client.close();
				}
				catch (IOException ignored)
				{
				}
			}
		}
	}
}

