/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.adaptor.http;

import java.io.IOException;
import javax.management.ObjectName;

import mx4j.adaptor.AdaptorServerSocketFactory;

/**
 * MBean of the HTTPAdapter
 *
 * @author <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version $Revision: 1.3 $
 */
public interface HttpAdaptorMBean
{
	/**
	 * Returns the port where the server is running on. Default is 80
	 *
	 * @return HTTPServer's port
	 */
	public int getPort();

	/**
	 * Sets the value of the server's port
	 *
	 * @param port the new port's value
	 */
	public void setPort(int port);

	/**
	 * Return the host name the server will be listening to. If null the server listen
	 * at the localhost
	 *
	 * @return the current hostname
	 */
	public String getHost();

	/**
	 * Sets the host name where the server will be listening
	 *
	 * @param host Server's host
	 */
	public void setHost(String host);

	/**
	 * Sets the object which create the server sockets
	 *
	 * @param factory the socket factory
	 */
	public void setSocketFactory(AdaptorServerSocketFactory factory);

	/**
	 * Sets the ObjectName of the factory for server sockets
	 *
	 * @param factoryName the ObjectName of the ServerSocket factory
	 */
	public void setSocketFactoryName(String factoryName);

	/**
	 * Sets the object which will post process the XML results. The last value
	 * set between the setProcessor and setPostProcessorName will be the
	 * valid one
	 *
	 * @param processor a processor object
	 */
	public void setProcessor(ProcessorMBean processor);

	/**
	 * Sets the object name which will post process the XML result. The last value
	 * set between the setProcessor and setPostProcessorName will be the
	 * valid one. The MBean will be verified to be of instance HttpPostProcessor
	 *
	 * @param processor a processor object
	 */
	public void setProcessorName(ObjectName processorName);

	/**
	 * Returns the object being used as a processor
	 *
	 * @return the processor object
	 */
	public ProcessorMBean getProcessor();

	/**
	 * Returns the ObjectName of the processor being used
	 *
	 * @return processor objectname
	 */
	public ObjectName getProcessorName();

	/**
	 * Authentication Method
	 *
	 * @return authentication method
	 */
	public String getAuthenticationMethod();

	/**
	 * Sets the Authentication Method.
	 *
	 * @param method none/basic/digest
	 */
	public void setAuthenticationMethod(String method);


	/**
	 * Adds an authorization
	 *
	 * @param username authorized username
	 * @param username authorized password
	 */
	public void addAuthorization(String username, String password);

	/**
	 * Indicates whether the server's running
	 *
	 * @param <CODE>true</CODE> if the server is working, <CODE>false</CODE> otherwise
	 */
	public boolean isActive();

	/**
	 * Starts the server
	 */
	public void start() throws IOException;

	/**
	 * Stops the server
	 */
	public void stop();

	/**
	 * Restarts the server. Useful when changing the Server parameters
	 */
	public void restart() throws IOException;
}
