/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.adaptor.http;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.JMException;
import javax.management.MBeanInfo;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanConstructorInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanParameterInfo;
import javax.management.MBeanNotificationInfo;
import javax.management.modelmbean.ModelMBeanInfo;
import javax.management.modelmbean.ModelMBeanAttributeInfo;
import javax.management.modelmbean.ModelMBeanConstructorInfo;
import javax.management.modelmbean.ModelMBeanOperationInfo;
import javax.management.modelmbean.ModelMBeanNotificationInfo;
import java.io.IOException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Text;
import javax.xml.parsers.DocumentBuilder;

/**
 * MBeanCommandProcessor, processes a request for getting data about a MBean
 *
 * @author <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version $Revision: 1.3 $
 */
class MBeanCommandProcessor implements HttpCommandProcessor
{
	private MBeanServer server;

	private DocumentBuilder builder;

	public MBeanCommandProcessor(MBeanServer server, DocumentBuilder builder)
	{
		this.server = server;
		this.builder = builder;
	}

	public Document executeRequest(HttpInputStream in) throws IOException, JMException
	{
		Document document = builder.newDocument();

		String name = in.getVariable("objectname");
		ObjectName objectName = null;
		// special case
		boolean validMBean = false;
		if (name != null)
		{
			objectName = new ObjectName(name);
			if (server.isRegistered(objectName))
			{
				validMBean = true;
			}
		}
		if (validMBean)
		{
			Element root = document.createElement("MBean");
			document.appendChild(root);

			MBeanInfo info = server.getMBeanInfo(objectName);
			root.setAttribute("description", info.getDescription());
			root.setAttribute("classname", info.getClassName());
			root.setAttribute("objectname", objectName.toString());
			boolean isModel = false;
			if (info instanceof ModelMBeanInfo)
			{
				isModel = true;
				root.setAttribute("model", "true");
			}
			if (HttpUtil.booleanVariableValue(in, "attributes", true)) {
				MBeanAttributeInfo[] attributes = info.getAttributes();
				if (attributes != null)
				{
					for (int i=0;i<attributes.length;i++)
					{
						Element attribute = document.createElement("Attribute");
						attribute.setAttribute("name", attributes[i].getName());
						attribute.setAttribute("type", attributes[i].getType());
						attribute.setAttribute("description", attributes[i].getDescription());
						if (attributes[i].isReadable() && attributes[i].isWritable())
						{
							attribute.setAttribute("availability", "RW");
						}
						if (attributes[i].isReadable() && !attributes[i].isWritable())
						{
							attribute.setAttribute("availability", "RO");
						}
						if (!attributes[i].isReadable() && attributes[i].isWritable())
						{
							attribute.setAttribute("availability", "WO");
						}
						try
						{
							Object attributeValue = server.getAttribute(objectName, attributes[i].getName());
							if (attributeValue != null)
							{
								attribute.setAttribute("value", attributeValue.toString());
							}
							else
							{
								attribute.setAttribute("value", "null");
							}
						}
						catch (JMException e)
						{
							attribute.setAttribute("value", e.getMessage());
						}
						root.appendChild(attribute);
					}
				}
			}
			if (HttpUtil.booleanVariableValue(in, "constructors", true))
			{
				MBeanConstructorInfo[] constructors = info.getConstructors();
				if (constructors != null)
				{
					for (int i=0;i<constructors.length;i++)
					{
						Element constructor = document.createElement("Constructor");
						constructor.setAttribute("name", constructors[i].getName());
						constructor.setAttribute("description", constructors[i].getDescription());
						addParameters(constructor, document, constructors[i].getSignature());
						root.appendChild(constructor);
					}
				}
			}
			if (HttpUtil.booleanVariableValue(in, "operations", true))
			{
				MBeanOperationInfo[] operations = info.getOperations();
				if (operations != null)
				{
					for (int i=0;i<operations.length;i++)
					{
						Element operation = document.createElement("Operation");
						operation.setAttribute("name", operations[i].getName());
						operation.setAttribute("description", operations[i].getDescription());
						operation.setAttribute("return", operations[i].getReturnType());
						switch (operations[i].getImpact())
						{
							case MBeanOperationInfo.UNKNOWN:
								operation.setAttribute("impact", "unknown");
								break;
							case MBeanOperationInfo.ACTION:
								operation.setAttribute("impact", "action");
								break;
							case MBeanOperationInfo.INFO:
								operation.setAttribute("impact", "info");
								break;
							case MBeanOperationInfo.ACTION_INFO:
								operation.setAttribute("impact", "action_info");
								break;
						}
						addParameters(operation, document, operations[i].getSignature());
						root.appendChild(operation);
					}
				}
			}
			if (HttpUtil.booleanVariableValue(in, "notifications", true))
			{
				MBeanNotificationInfo[] notifications = info.getNotifications();
				if (notifications != null)
				{
					for (int i=0;i<notifications.length;i++)
					{
						Element notification = document.createElement("Notification");
						notification.setAttribute("name", notifications[i].getName());
						notification.setAttribute("description", notifications[i].getDescription());
						String[] types = notifications[i].getNotifTypes();
						for (int j=0;j<types.length;j++)
						{
							Element type =  document.createElement("Type");
							type.setAttribute("name", types[j]);
							notification.appendChild(type);
						}
						root.appendChild(notification);
					}
				}
			}
		}
		return document;
	}

	protected void addParameters(Element node, Document document, MBeanParameterInfo[] parameters)
	{
		for (int j=0;j<parameters.length;j++)
		{
			Element parameter = document.createElement("Parameter");
			parameter.setAttribute("name", parameters[j].getName());
			parameter.setAttribute("description", parameters[j].getDescription());
			parameter.setAttribute("type", parameters[j].getType());
			// add id since order is relevant
			parameter.setAttribute("id", ""+j);
			node.appendChild(parameter);
		 }
	}
}
