/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */
package mx4j.adaptor.http;

import java.io.IOException;
import org.w3c.dom.Document;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.io.File;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedInputStream;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.net.URL;
import java.net.URLClassLoader;
import java.net.URLConnection;
import java.net.MalformedURLException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.Templates;
import javax.xml.transform.Source;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.URIResolver;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.transform.stream.StreamResult;
import mx4j.log.Log;
import mx4j.log.Logger;
import javax.management.RuntimeMBeanException;

/**
 * XSLTPostProcessor pass the document through an XSLT transformation
 *
 * @author <a href="mailto:tibu@users.sourceforge.net">Carlos Quiroz</a>
 * @version $Revision: 1.2 $
 */
public class XSLTProcessor
	implements ProcessorMBean, XSLTProcessorMBean, URIResolver
{
	TransformerFactory factory = null;

	private Map templatesCache = new HashMap();

	private static final String ENCODING = "UTF-8";

	private String path = "mx4j/adaptor/http/xsl";

	private File root = null;

	private Map mimeTypes = new HashMap();

	private Logger log = null;

	/** Indicated whether the file are read from a file */
	private boolean useJar = true;

	private boolean useCache = true;

	private ClassLoader targetClassLoader = ClassLoader.getSystemClassLoader();

	public XSLTProcessor()
	{
		factory = TransformerFactory.newInstance();
		factory.setURIResolver(this);
		mimeTypes.put(".gif", "image/gif");
		mimeTypes.put(".jpg", "image/jpg");
		mimeTypes.put(".png", "image/png");
		mimeTypes.put(".html", "text/html");
		mimeTypes.put(".htm", "text/html");
		mimeTypes.put(".css", "text/css");
		log = Log.getLogger("XSLTProcessor");
	}

	public void writeResponse(HttpOutputStream out, HttpInputStream in, Document document) throws IOException
	{
		out.setCode(HttpConstants.STATUS_OKAY);
		out.setHeader("Content-Type", "text/html");
		out.sendHeaders();
		Transformer transformer = null;
		String path = preProcess(in.getPath());

		if (in.getVariable("template") != null)
		{
			transformer = createTransformer(in.getVariable("template") + ".xsl");
		}
		else
		{
			transformer = createTransformer(path + ".xsl");
		}

		if (transformer != null)
		{
			// The variables are passed to the XSLT as (param.name, value)
			Map variables = in.getVariables();
			Iterator j = variables.keySet().iterator();
			while (j.hasNext())
			{
				String key = (String)j.next();
				String value = (String)variables.get(key);
				transformer.setParameter("request." + key, value);
			}
			try
			{
				ByteArrayOutputStream output = new ByteArrayOutputStream();
				log.trace("transforming " + path);
				transformer.transform(new DOMSource(document), new StreamResult(output));
				output.writeTo(out);
			}
			catch (TransformerException e)
			{
				log.error("Transformation exception ", e);
			}
		}
		else
		{
			log.warn("Transformer for path " + path + " not found");
		}
	}

	protected Transformer createTransformer(String path)
	{
		Transformer transformer = null;
		try
		{
			if (useCache && templatesCache.containsKey(path))
			{
				transformer = ((Templates)templatesCache.get(path)).newTransformer();
			}
			else
			{
				InputStream file = getInputStream(path);
				if (file != null)
				{
					log.info("Creating template for path " + path);
					Templates template = factory.newTemplates(new StreamSource(file));
					transformer = template.newTransformer();
					if (useCache)
					{
						templatesCache.put(path, template);
					}
				}
				else
				{
					log.warn("template for path " + path + " not found");
				}
			}
		}
		catch (TransformerConfigurationException e)
		{
			log.error("Exception during template construction", e);
		}
		return transformer;
	}

	public void writeError(HttpOutputStream out, HttpInputStream in, Exception e) throws IOException
	{
		Exception t = e;
		if (e instanceof RuntimeMBeanException)
		{
			t = ((RuntimeMBeanException)e).getTargetException();
		}
		log.info("Processing error " + t.getMessage());
		if (t instanceof HttpException)
		{
			out.setCode(((HttpException)e).getCode());
			out.setHeader("Content-Type", "text/html");
			out.sendHeaders();
			// hardcoded dir :-P
			Transformer transformer = createTransformer("error.xsl");
			Document doc = ((HttpException)e).getResponseDoc();
			if (doc != null)
			{
				try
				{
					ByteArrayOutputStream output = new ByteArrayOutputStream();
					transformer.transform(new DOMSource(doc), new StreamResult(output));
					output.writeTo(out);
				}
				catch (TransformerException ex)
				{
					log.error("Exception during error output", ex);
				}
			}
		}
		else
		{
			out.setCode(HttpConstants.STATUS_INTERNAL_ERROR);
			out.setHeader("Content-Type", "text/html");
			out.sendHeaders();
		}
	}

	public String preProcess(String path) {
		if (path.equals("/"))
		{
			path = "/serverbydomain";
		}
		return path;
	}

	public String notFoundElement(String path, HttpOutputStream out, HttpInputStream in) throws IOException, HttpException
	{
		File file = new File(this.path, path);
		log.info("Processing file request " + file);
		boolean found = false;
		String name = file.getName();
		int extensionIndex = name.lastIndexOf('.');
		String mime = null;
		if (extensionIndex < 0)
		{
			log.warn("Filename has no extensions " + file.toString());
		}
		else
		{
			String extension = name.substring(extensionIndex, name.length());
			if (mimeTypes.containsKey(extension))
			{
				mime = (String)mimeTypes.get(extension);
				found = true;
			}
			else
			{
				log.warn("MIME type not found " + extension);
			}
		}
		if (found)
		{
			try
			{
				log.debug("Reading file");
				BufferedInputStream fileIn = new BufferedInputStream(getInputStream(path));
				ByteArrayOutputStream outArray = new ByteArrayOutputStream();
				BufferedOutputStream outBuffer = new BufferedOutputStream(outArray);
				int piece = 0;
				while ((piece = fileIn.read()) >= 0)
				{
					outBuffer.write(piece);
				}
				outBuffer.flush();
				out.setCode(HttpConstants.STATUS_OKAY);
				out.setHeader("Content-type", mime);
				out.sendHeaders();
				log.debug("File output " + mime);
				outArray.writeTo(out);
				fileIn.close();
			}
			catch (Exception e)
			{
				e.printStackTrace();
				found = false;
			}
		}
		if (!found)
		{
			throw new HttpException(HttpConstants.STATUS_NOT_FOUND, "file not found");
		}
		return null;
	}

	protected InputStream getInputStream(String path)
	{
		InputStream file = null;
		if (!useJar)
		{
			try
			{
				// load from a dir
				file = new FileInputStream(new File(this.root, path));
			}
			catch (FileNotFoundException e)
			{
				log.error("File not found", e);
			}
		}
		else
		{
			// load from a jar
			String targetFile = this.path;
			// workaround, should tought of somehting better
			if (path.startsWith("/"))
			{
				targetFile += path;
			} else {
				targetFile += "/" + path;
			}
			if (root != null)
			{
				file = targetClassLoader.getResourceAsStream(targetFile);
			}
			if (file == null)
			{
				file = getClass().getClassLoader().getResourceAsStream(targetFile);
			}
		}

		return file;
	}

	public Source resolve(String href, String base)
	{
		return new StreamSource(getInputStream(href));
	}

	public void setFile(String file)
	{
		if (file != null)
		{
			File target = new File(file);
			if (!target.exists())
			{
				log.warn("Target file does not exist, defaulting to previous");
				return;
			}
			if (target.isDirectory())
			{
				useJar = false;
				log.info("Using " + file + " as the root dir");
				this.root = target;
				return;
			}
			if (target.isFile() && (target.getName().endsWith(".jar") ||
				(target.getName().endsWith(".zip"))))
			{
				try {
					URL url = target.toURL();
					targetClassLoader = new URLClassLoader(new URL[] {url});
					log.info("Using compressed file " + url + " as the root file");
					this.root = target;
					useJar = true;
				} catch (MalformedURLException e) {
					log.warn("Unable to create class loader", e);
				}
			}
			else
			{
				log.warn("Target file does not exist, defaulting to previous");
			}
		}
	}

	public String getFile()
	{
		return (root != null)?root.getName():null;
	}

	public String getPathInJar()
	{
		return path;
	}

	public void setPathInJar(String path)
	{
		this.path = path;
	}

	public boolean isUseJar()
	{
		return useJar;
	}

	public boolean isUsePath()
	{
		return !useJar;
	}

	public void addMimeType(String extension, String type)
	{
		if (extension != null && type != null)
		{
			log.info("Added MIME type " + type + " for extension " + extension);
			mimeTypes.put(extension, type);
		}
	}

	public void setUseCache(boolean useCache)
	{
		this.useCache = useCache;
	}

	public boolean isUseCache()
	{
		return useCache;
	}

	public String getName()
	{
		return "XSLT Processor";
	}
}

