/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.adaptor.rmi;

import java.util.Properties;
import java.util.ArrayList;
import java.util.List;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.net.InetAddress;
import java.net.UnknownHostException;

import javax.management.MBeanRegistration;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import mx4j.adaptor.interceptor.Interceptor;

/**
 * The RMI adaptor MBean
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public abstract class RMIAdaptor implements RMIAdaptorMBean, MBeanRegistration
{
	private MBeanServer m_server;
	private ObjectName m_objectName;
	private String m_jndiName;
	private Properties m_namingProperties = new Properties();
	private ArrayList m_interceptors = new ArrayList();

	public ObjectName preRegister(MBeanServer server, ObjectName name) throws Exception
	{
		m_server = server;
		m_objectName = name == null ? getDefaultObjectName() : name;
		return m_objectName;
	}

	public void postRegister(Boolean registrationDone)
	{
	}

	public void preDeregister() throws Exception
	{
		stop();
	}

	public void postDeregister()
	{
	}

	public String getJNDIName()
	{
		return m_jndiName;
	}

	public void setJNDIName(String name)
	{
		if (!isRunning())
		{
			m_jndiName = name;
		}
		else
		{
			throw new IllegalStateException("Cannot perform this operation while running");
		}
	}

	public void putNamingProperty(Object property, Object value)
	{
		if (!isRunning())
		{
			m_namingProperties.put(property, value);
		}
		else
		{
			throw new IllegalStateException("Cannot perform this operation while running");
		}
	}

	public void clearNamingProperties()
	{
		if (!isRunning())
		{
			m_namingProperties.clear();
		}
		else
		{
			throw new IllegalStateException("Cannot perform this operation while running");
		}
	}

	public String getHostName()
	{
		try
		{
			return InetAddress.getLocalHost().getHostName();
		}
		catch (UnknownHostException ignored)
		{
			return "localhost";
		}
	}

	/**
	 * Adds an interceptor to this RMI adaptor
	 */
	public void addInterceptor(Interceptor interceptor)
	{
		if (interceptor != null) {m_interceptors.add(interceptor);}
	}

	/**
	 * Sets the target MBeanServer in case this adaptor is not registered with it
	 */
	public void setMBeanServer(MBeanServer server)
	{
		m_server = server;
	}

	protected List getInterceptors()
	{
		return m_interceptors;
	}

	public Properties getNamingProperties()
	{
		return (Properties)m_namingProperties.clone();
	}

	protected MBeanServer getMBeanServer()
	{
		return m_server;
	}

	protected ObjectName getObjectName()
	{
		return m_objectName;
	}

	protected abstract ObjectName getDefaultObjectName() throws MalformedObjectNameException;

	protected void bind(Object adaptor) throws NamingException
	{
		InitialContext context = null;
		Properties properties = getNamingProperties();
		if (properties.size() != 0)
		{
			context = new InitialContext(properties);
		}
		else
		{
			context = new InitialContext();
			Hashtable env = context.getEnvironment();
			for (Iterator i = env.entrySet().iterator(); i.hasNext();)
			{
				Map.Entry entry = (Map.Entry)i.next();
				putNamingProperty(entry.getKey(), entry.getValue());
			}
		}
		context.bind(getJNDIName(), adaptor);
	}

	protected void unbind() throws NamingException
	{
		InitialContext context = null;
		Properties properties = getNamingProperties();
		if (properties.size() != 0)
		{
			context = new InitialContext(properties);
		}
		else
		{
			context = new InitialContext();
		}
		context.unbind(getJNDIName());
	}
}
