/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.adaptor.rmi;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.rmi.server.RemoteObject;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.List;
import java.util.Arrays;
import java.util.ArrayList;

import javax.management.MBeanServer;
import javax.management.ObjectName;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanRegistrationException;

import mx4j.adaptor.interceptor.AdaptorInterceptor;
import mx4j.adaptor.interceptor.Interceptor;
import mx4j.adaptor.interceptor.Invocation;
import mx4j.adaptor.interceptor.InvocationResult;
import mx4j.util.Utils;
import mx4j.util.MethodTernaryTree;
import mx4j.log.Logger;
import mx4j.log.Log;

/**
 * The remote object that forwards call to the MBeanServer.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public class RMIProtocolAdaptorImpl extends RemoteObject implements RMIProtocolAdaptor
{
	private MBeanServer m_server;
	private ObjectName m_name;
	private Interceptor m_headInterceptor;
	private String m_hostName;
	private String m_hostAddress;
	private ArrayList m_interceptorNames = new ArrayList();

	public RMIProtocolAdaptorImpl(MBeanServer server, ObjectName name, List interceptors)
	{
		m_server = server;
		m_name = name;

		if (interceptors == null || interceptors.size() == 0)
		{
			m_headInterceptor = new InvokerAdaptorInterceptor();
		}
		else
		{
			Logger logger = getLogger();
			if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Interceptors are: " + interceptors);}

			for (int i = 1; i < interceptors.size(); ++i)
			{
				Interceptor previous = (Interceptor)interceptors.get(i - 1);
				Interceptor next = (Interceptor)interceptors.get(i);
				previous.setNext(next);
				registerInterceptor(previous);
			}

			Interceptor last = (Interceptor)interceptors.get(interceptors.size() - 1);
			last.setNext(new InvokerAdaptorInterceptor());
			registerInterceptor(last);

			m_headInterceptor = (Interceptor)interceptors.get(0);
		}
	}

	private void registerInterceptor(Interceptor interceptor)
	{
		if (m_name == null)
		{
			// Means that this interceptor has not been registered, don't register the interceptors
			return;
		}

		Logger logger = getLogger();

		try
		{
			ObjectName interceptorObjectName = null;

			if (interceptor instanceof AdaptorInterceptor)
			{
				AdaptorInterceptor i = (AdaptorInterceptor)interceptor;
				interceptorObjectName = i.getObjectName();
			}

			if (interceptorObjectName == null)
			{
				// Deduct the interceptor name from the adaptor name

				if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Adaptor ObjectName: " + m_name);}

				String domain = m_name.getDomain();
				Hashtable props = m_name.getKeyPropertyList();
				props.put("interceptor", interceptor.getType());
				interceptorObjectName = new ObjectName(domain, props);
			}

			if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Adaptor interceptor ObjectName: " + interceptorObjectName);}
			m_server.registerMBean(interceptor, interceptorObjectName);
			m_interceptorNames.add(interceptorObjectName);
			if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Adaptor interceptor registered successfully");}
		}
		catch (Exception x)
		{
			if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Cannot register interceptor", x);}
		}
	}

	public void unregisterInterceptors()
	{
		Logger logger = getLogger();

		for (int i = 0; i < m_interceptorNames.size(); ++i)
		{
			ObjectName name = (ObjectName)m_interceptorNames.get(i);
			if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Unregistering adaptor interceptor with ObjectName: " + name);}
			try
			{
				m_server.unregisterMBean(name);
				if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Adaptor interceptor unregistered successfully");}
			}
			catch (Exception x)
			{
				if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Cannot unregister interceptor", x);}
			}
		}
	}

	public InvocationResult invoke(Invocation invocation) throws Exception
	{
		return m_headInterceptor.invoke(invocation);
	}

	public String getHostName()
	{
		if (m_hostName == null)
		{
			try
			{
				InetAddress host = InetAddress.getLocalHost();
				m_hostName = host.getHostName();
			}
			catch (UnknownHostException ignored)
			{
			}
		}
		return m_hostName;
	}

	public String getHostAddress()
	{
		if (m_hostAddress == null)
		{
			try
			{
				InetAddress host = InetAddress.getLocalHost();
				m_hostAddress = host.getHostAddress();
			}
			catch (UnknownHostException ignored)
			{
			}
		}
		return m_hostAddress;
	}

	private Logger getLogger()
	{
		return Log.getLogger("mx4j.implementation." + getClass().getName());
	}

	public final class InvokerAdaptorInterceptor extends AdaptorInterceptor
	{
		private MethodTernaryTree m_methods = new MethodTernaryTree();

		public InvokerAdaptorInterceptor()
		{
			registerInterceptor(this);
		}

		public String getType()
		{
			return "invoker";
		}

		/**
		 * Cannot change the status of this attribute, this interceptor is always enabled
		 */
		public void setEnabled(boolean value) {}

		/**
		 * This interceptor is always enabled
		 */
		public boolean isEnabled()
		{
			return true;
		}

		protected InvocationResult doInvoke(Invocation invocation) throws Exception
		{
			// Here I have to call the MBeanServer

			Logger logger = this.getLogger();
			boolean debug = logger.isEnabledFor(Logger.DEBUG);
			if (debug) {logger.debug("Last interceptor, invoking MBeanServer");}

			// Extract method name from invocation
			String name = invocation.getMethodName();
			if (debug) {logger.debug("Method name is: " + name);}

			// Extract signature from invocation
			String[] signature = invocation.getSignature();
			Class[] params = Utils.loadClasses(Thread.currentThread().getContextClassLoader(), signature);
			if (debug) {logger.debug("Signature is: " + Arrays.asList(signature));}

			// Extract arguments
			Object[] args = invocation.getArguments();
			if (debug) {logger.debug("Arguments are: " + Arrays.asList(args));}

			try
			{
				Object result = null;
				if (isMBeanServerInvocation(name, params))
				{
					if (m_server == null) {throw new IllegalStateException("Target MBeanServer is not set. Either call setMBeanServer or register this MBean");}
					Method m = getMBeanServerMethod(name, signature, params);
               result = invokeMBeanServerMethod(m, args);
				}
				else
				{
					// TODO: cache also these methods
					Method m = RMIProtocolAdaptorImpl.this.getClass().getMethod(name, params);
					result = m.invoke(RMIProtocolAdaptorImpl.this, args);
				}

				InvocationResult invocationResult = new InvocationResult();
				invocationResult.setResult(result);

				return invocationResult;
			}
			catch (InvocationTargetException x)
			{
				// Unwrap invocation target exceptions
				Throwable t = x.getTargetException();
				if (t instanceof Exception) {throw (Exception)t;}
				else {throw (Error)t;}
			}
		}

		private Method getMBeanServerMethod(String name, String[] signature, Class[] params) throws NoSuchMethodException
		{
			// Double checked locking does not work in Java :(

			synchronized (m_methods)
			{
				Method m = (Method)m_methods.get(name, signature);
				if (m == null)
				{
               try
               {
                  m = MBeanServer.class.getMethod(name, params);
               }
               catch (NoSuchMethodException x)
               {
                  // The only problem that may happen is that the RemoteMBeanServer
                  // has a removeNotificationListener with a different signature
                  // Waiting that the RI will update to a decent signature for
                  // adding and removing listeners
                  if ("removeNotificationListener".equals(name))
                  {
                     Class[] newParams = new Class[] {params[0], params[1]};
                     m = MBeanServer.class.getMethod(name, newParams);
                  }
                  else
                  {
                     throw x;
                  }
               }
               m_methods.put(name, signature, m);
				}
				return m;
			}
		}

      private Object invokeMBeanServerMethod(Method method, Object[] params) throws Exception
      {
         if ("removeNotificationListener".equals(method.getName()))
         {
            Object[] newParams = new Object[] {params[0], params[1]};
            return method.invoke(m_server, newParams);
         }
         else
         {
            return method.invoke(m_server, params);
         }
      }

		private boolean isMBeanServerInvocation(String name, Class[] params)
		{
			if (name.equals("getHostName") || name.equals("getHostAddress"))
			{
				return false;
			}
			return true;
		}
	}
}
