/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.connector.rmi;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.rmi.RemoteException;
import java.rmi.MarshalledObject;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Set;
import java.io.IOException;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.NotCompliantMBeanException;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.QueryExp;
import javax.management.ReflectionException;
import javax.management.RuntimeOperationsException;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import mx4j.adaptor.interceptor.Invocation;
import mx4j.adaptor.interceptor.InvocationContext;
import mx4j.adaptor.interceptor.InvocationResult;
import mx4j.adaptor.rmi.RMIProtocolAdaptor;
import mx4j.connector.ConnectorException;
import mx4j.connector.JMXConnector;
import mx4j.connector.RemoteMBeanServer;

/**
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.5 $
 */
public abstract class RMIConnector implements JMXConnector
{
	private RMIProtocolAdaptor m_adaptor;
	private RemoteMBeanServer m_server;
	private InvocationContext m_context;

	public void connect(String jndiName, Hashtable properties) throws Exception
	{
		if (jndiName == null)
		{
			throw new NamingException("Null JNDI name");
		}

		InitialContext ctx = new InitialContext(properties);
		Object object = ctx.lookup(jndiName);

		m_adaptor = narrow(object);

		m_server = new RemoteMBeanServerImpl();
	}

	public RemoteMBeanServer getRemoteMBeanServer()
	{
		if (m_server == null)
		{
			throw new IllegalStateException("connect() has not been called");
		}
		return m_server;
	}

	protected abstract RMIProtocolAdaptor narrow(Object object);

	public void setInvocationContext(InvocationContext context)
	{
		m_context = context;
	}

	public String getRemoteHostName()
	{
		try
		{
			return (String)invoke("getHostName", new String[0], new Object[0]);
		}
		catch (ConnectorException x)
		{
			throw x;
		}
		catch (RuntimeException x)
		{
			throw x;
		}
		catch (Exception ignored)
		{
			ignored.printStackTrace();
			return null;
		}
	}

	public String getRemoteHostAddress()
	{
		try
		{
			return (String)invoke("getHostAddress", new String[0], new Object[0]);
		}
		catch (ConnectorException x)
		{
			throw x;
		}
		catch (RuntimeException x)
		{
			throw x;
		}
		catch (Exception ignored)
		{
			ignored.printStackTrace();
			return null;
		}
	}

	protected abstract NotificationListener createRemoteNotificationListener(NotificationListener listener) throws RemoteException;

	protected abstract NotificationFilter createRemoteNotificationFilter(NotificationFilter listener) throws RemoteException;

	protected Object invoke(String methodName, String[] signature, Object[] arguments) throws Exception
	{
		Invocation invocation = new Invocation();
		invocation.setMethodName(methodName);
		invocation.setSignature(signature);
		invocation.setArguments(arguments);
		invocation.setInvocationContext(m_context);

		try
		{
			InvocationResult result = m_adaptor.invoke(invocation);
			return result.getResult();
		}
		catch (RemoteException x)
		{
			// A runtime exception that wraps remotes
			throw new ConnectorException(x);
		}
	}

	private class RemoteMBeanServerImpl implements RemoteMBeanServer
	{
		private RemoteMBeanServer m_proxy;
		private HashMap m_listeners = new HashMap();
		private HashMap m_filters = new HashMap();
      private HashMap m_handbacks = new HashMap();

		private RemoteMBeanServerImpl()
		{
			m_proxy = (RemoteMBeanServer)Proxy.newProxyInstance(RemoteMBeanServer.class.getClassLoader(), new Class[]{RemoteMBeanServer.class}, new Handler());
		}

		public void addNotificationListener(ObjectName observed, NotificationListener listener, NotificationFilter filter, Object handback)
				throws InstanceNotFoundException
		{
			if (listener == null)
			{
				throw new RuntimeOperationsException(new IllegalArgumentException("NotificationListener cannot be null"));
			}

         NotificationListener l = null;
         NotificationFilter f = null;
         Object h = null;
         synchronized (this)
         {
            // Handle the listener
            l = findListener(listener);
            // Handle the filter
            f = findFilter(filter);
            // Handle the handback
            h = findHandback(handback);
         }

         // Invoke the remote server
         m_proxy.addNotificationListener(observed, l, f, h);

         synchronized (this)
         {
            // Everything went ok, register this listener-filter-handback trio
            // Overwrite in case of concurrent threads

            m_listeners.put(listener, l);
            if (filter != null) {m_filters.put(filter, f);}
            if (handback != null) {m_handbacks.put(handback, h);}
         }
		}

		public void removeNotificationListener(ObjectName observed, NotificationListener listener, NotificationFilter filter, Object handback)
				throws InstanceNotFoundException, ListenerNotFoundException
		{
			if (listener == null)
			{
				throw new RuntimeOperationsException(new IllegalArgumentException("NotificationListener cannot be null"));
			}

         NotificationListener l = null;
         NotificationFilter f = null;
         Object h = null;
         synchronized (this)
         {
            l = (NotificationListener)m_listeners.get(listener);
            f = (NotificationFilter)m_filters.get(filter);
            h = m_handbacks.get(handback);
         }

         // Invoke the remote server
         m_proxy.removeNotificationListener(observed, l, f, h);

         synchronized (this)
         {
            // Everything went ok, remove this listener-filter-handback trio
            m_listeners.remove(listener);
            if (filter != null) {m_filters.remove(filter);}
            if (handback != null) {m_handbacks.remove(handback);}
         }
		}

		public ObjectInstance createMBean(String className, ObjectName objectName)
				throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException
		{
			return m_proxy.createMBean(className, objectName);
		}

		public ObjectInstance createMBean(String className, ObjectName objectName, ObjectName loaderName)
				throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException, InstanceNotFoundException
		{
			return m_proxy.createMBean(className, objectName, loaderName);
		}

		public ObjectInstance createMBean(String className, ObjectName objectName, Object[] args, String[] parameters)
				throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException
		{
			return m_proxy.createMBean(className, objectName, args, parameters);
		}

		public ObjectInstance createMBean(String className, ObjectName objectName, ObjectName loaderName, Object[] args, String[] parameters)
				throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException, InstanceNotFoundException
		{
			return m_proxy.createMBean(className, objectName, loaderName, args, parameters);
		}

		public ObjectInstance registerMBean(Object mbean, ObjectName objectName)
				throws InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException
		{
			return m_proxy.registerMBean(mbean, objectName);
		}

		public void unregisterMBean(ObjectName objectName)
				throws InstanceNotFoundException, MBeanRegistrationException
		{
			m_proxy.unregisterMBean(objectName);
		}

		public Object getAttribute(ObjectName objectName, String attribute)
				throws MBeanException, AttributeNotFoundException, InstanceNotFoundException, ReflectionException
		{
			return m_proxy.getAttribute(objectName, attribute);
		}

		public void setAttribute(ObjectName objectName, Attribute attribute)
				throws InstanceNotFoundException, AttributeNotFoundException, InvalidAttributeValueException, MBeanException, ReflectionException
		{
			m_proxy.setAttribute(objectName, attribute);
		}

		public AttributeList getAttributes(ObjectName objectName, String[] attributes)
				throws InstanceNotFoundException, ReflectionException
		{
			return m_proxy.getAttributes(objectName, attributes);
		}

		public AttributeList setAttributes(ObjectName objectName, AttributeList attributes)
				throws InstanceNotFoundException, ReflectionException
		{
			return m_proxy.setAttributes(objectName, attributes);
		}

		public Object invoke(ObjectName objectName, String methodName, Object[] args, String[] parameters)
				throws InstanceNotFoundException, MBeanException, ReflectionException
		{
			return m_proxy.invoke(objectName, methodName, args, parameters);
		}

		public String getDefaultDomain()
		{
			return m_proxy.getDefaultDomain();
		}

		public Integer getMBeanCount()
		{
			return m_proxy.getMBeanCount();
		}

		public boolean isRegistered(ObjectName objectname)
		{
			return m_proxy.isRegistered(objectname);
		}

		public MBeanInfo getMBeanInfo(ObjectName objectName)
				throws InstanceNotFoundException, IntrospectionException, ReflectionException
		{
			return m_proxy.getMBeanInfo(objectName);
		}

		public ObjectInstance getObjectInstance(ObjectName objectName)
				throws InstanceNotFoundException
		{
			return m_proxy.getObjectInstance(objectName);
		}

		public boolean isInstanceOf(ObjectName objectName, String className)
				throws InstanceNotFoundException
		{
			return m_proxy.isInstanceOf(objectName, className);
		}

		public Set queryMBeans(ObjectName patternName, QueryExp filter)
		{
			return m_proxy.queryMBeans(patternName, filter);
		}

		public Set queryNames(ObjectName patternName, QueryExp filter)
		{
			return m_proxy.queryNames(patternName, filter);
		}

		private NotificationListener findListener(NotificationListener listener)
		{
         // First, lookup in the map
			NotificationListener l = (NotificationListener)m_listeners.get(listener);
         if (l != null) {return l;}

         // Check if user provided a specific remote listener, otherwise go default
         if (listener instanceof RemoteNotificationListenerSupport)
         {
            return listener;
         }
         else
         {
            try
            {
               return createRemoteNotificationListener(listener);
            }
            catch (RemoteException x)
            {
               throw new ConnectorException(x);
            }
         }
		}

		private NotificationFilter findFilter(NotificationFilter filter)
		{
         if (filter == null) {return null;}

			NotificationFilter f = (NotificationFilter)m_filters.get(filter);
         if (f != null) {return f;}

         // Check if user provided a specific remote filter, otherwise go default
         if (filter instanceof RemoteNotificationFilterSupport)
         {
            return filter;
         }
         else
         {
            try
            {
               return createRemoteNotificationFilter(filter);
            }
            catch (RemoteException x)
            {
               throw new ConnectorException(x);
            }
         }
		}

      private Object findHandback(Object handback)
      {
         if (handback == null) {return null;}

         Object h = m_handbacks.get(handback);
         if (h != null) {return h;}

         // Replace the handback with a serializable object
         try
         {
            return new MarshalledObject(handback);
         }
         catch (IOException x)
         {
            throw new RuntimeOperationsException(new IllegalArgumentException("Handback is not serializable"));
         }
      }

		private class Handler implements InvocationHandler
		{
			public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
			{
				// Needed for no-parameter methods
				if (args == null)
				{
					args = new Object[0];
				}

				int length = method.getParameterTypes().length;
				String[] signature = new String[length];
				for (int i = 0; i < length; ++i)
				{
					signature[i] = method.getParameterTypes()[i].getName();
				}

				return RMIConnector.this.invoke(method.getName(), signature, args);
			}
		}
	}
}
