/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server;

import java.security.Permission;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.StringTokenizer;

import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;

import mx4j.util.Utils;

/**
 * Permission that control access to MBean instances registered in the MBeanServer. <p>
 * The MBeanPermission contains a target name and a comma separated list of target actions.
 * The target name is composed by:
 * <ul>
 * <li> the class name of the MBean, as returned by
 * {@link javax.management.MBeanInfo#getClassName MBeanInfo.getClassName()} </li>
 * <li> the pound character '#' </li>
 * <li> the attribute name or the operation name </li>
 * <li> the object name of the MBean inclosed in squared brackets </li>
 * </ul>
 * The wildcard '*' may be used to specify packages, classes or methods as a whole. <br>
 * The action's list is the following:
 * <ul>
 * <li> addListener </li>
 * <li> removeListener </li>
 * <li> getMBeanInfo </li>
 * <li> getAttribute </li>
 * <li> setAttribute </li>
 * <li> invokeOperation </li>
 * <li> newMBean </li>
 * <li> registerMBean </li>
 * <li> unregisterMBean </li>
 * </ul>
 * The wildcard '*' indicates all actions. <br>
 * An example of policy file is the following:
 * <pre>
 * grant codebase my-jmx-application.jar
 * {
 *    permission mx4j.server.MBeanPermission "mx4j.tools.naming.NamingService", "newMBean, registerMBean, unregisterMBean";
 *    permission mx4j.server.MBeanPermission "mx4j.tools.naming.NamingService#start", "invokeOperation";
 *    permission mx4j.server.MBeanPermission "mx4j.tools.naming.NamingService#stop", "invokeOperation";
 * }
 * </pre>
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.4 $
 */
public final class MBeanPermission extends Permission
{
	private String m_actions;
	private transient int m_hash;
	private transient String m_targetName;
	private transient ObjectName m_objectName;
	private transient ArrayList m_actionsList;

	public MBeanPermission(String name)
	{
		this(name, null);
	}

	public MBeanPermission(String name, String actions)
	{
		super(name);
		m_actions = actions;

		parse(name, actions);
	}

	private void parse(String name, String actions)
	{
		m_targetName = parseTargetName(name);
		m_objectName = parseObjectName(name);
		m_actionsList = parseActions(actions);
	}

	public int hashCode()
	{
		if (m_hash == 0)
		{
			m_hash = computeHash();
		}
		return m_hash;
	}

	public boolean equals(Object obj)
	{
		if (obj == null) {return false;}
		if (obj == this) {return true;}

		try
		{
			// Must have the same target name, and the same action list
			MBeanPermission other =  (MBeanPermission)obj;

			if (!getName().equals(other.getName())) {return false;}

			ArrayList thisActions = getActionsList();
			ArrayList otherActions = other.getActionsList();
			if (thisActions.containsAll(otherActions) && otherActions.containsAll(thisActions)) {return true;}
		}
		catch (ClassCastException ignored) {}
		return false;
	}

	public String getActions()
	{
		return m_actions;
	}

	public String getTargetName()
	{
		return m_targetName;
	}

	public ObjectName getObjectName()
	{
		return m_objectName;
	}

	private ArrayList getActionsList()
	{
		return m_actionsList;
	}

	public boolean implies(Permission p)
	{
		if (!(p instanceof MBeanPermission)) {return false;}

		MBeanPermission permission = (MBeanPermission)p;
		if (!impliesTargetName(permission)) {return false;}
		if (!impliesObjectName(permission)) {return false;}
		if (!impliesActions(permission)) {return false;}
		return true;
	}

	private boolean impliesTargetName(MBeanPermission p)
	{
		String thisName = getTargetName();
		String otherName = p.getTargetName();

		boolean thisWildcard = thisName.endsWith(".*") || thisName.equals("*");
		if (thisWildcard) {thisName = thisName.substring(0, thisName.length() - 1);}
		boolean otherWildcard = otherName.endsWith(".*") || otherName.equals("*");
		if (otherWildcard) {otherName = otherName.substring(0, otherName.length() - 1);}

		if (thisWildcard)
		{
			if (otherWildcard)
			{
				return otherName.startsWith(thisName);
			}
			else
			{
				return otherName.length() > thisName.length() && otherName.startsWith(thisName);
			}
		}
		else
		{
			if (otherWildcard)
			{
				return false;
			}
			else
			{
				return thisName.equals(otherName);
			}
		}
	}

	private boolean impliesObjectName(MBeanPermission p)
	{
		ObjectName thisName = getObjectName();
		if (thisName == null)
		{
			try	{thisName = new ObjectName("*:*");}
			catch (MalformedObjectNameException ignored) {}
		}
		ObjectName otherName = p.getObjectName();
		if (otherName == null)
		{
			try	{otherName = new ObjectName("*:*");}
			catch (MalformedObjectNameException ignored) {}
		}

		// See if domains match
		String thisDomain = thisName.getDomain();
		boolean thisDomainIsPattern = thisDomain.indexOf('*') >= 0 || thisDomain.indexOf('?') >= 0;
		String otherDomain = otherName.getDomain();
		boolean otherDomainIsPattern = otherDomain.indexOf('*') >= 0 || otherDomain.indexOf('?') >= 0;

		if (!thisDomainIsPattern && otherDomainIsPattern) {return false;}
		if (!thisDomainIsPattern && !otherDomainIsPattern && !thisDomain.equals(otherDomain)) {return false;}
		if (!Utils.wildcardMatch(thisDomain, otherDomain)) {return false;}

		// Here domains match, see properties
		Hashtable thisProperties = thisName.getKeyPropertyList();
		boolean thisPropertiesIsPattern = thisName.isPropertyPattern();
		Hashtable otherProperties = otherName.getKeyPropertyList();
		boolean otherPropertiesIsPattern = otherName.isPropertyPattern();

		if (!thisPropertiesIsPattern && otherPropertiesIsPattern) {return false;}
		if (!thisPropertiesIsPattern && !otherPropertiesIsPattern && !thisProperties.equals(otherProperties)) {return false;}
      if (thisPropertiesIsPattern && !otherProperties.entrySet().containsAll(thisProperties.entrySet())) {return false;}

		return true;
	}

	private boolean impliesActions(MBeanPermission p)
	{
		ArrayList otherActions = p.getActionsList();
		return getActionsList().containsAll(otherActions);
	}

	private String parseTargetName(String name)
	{
		if (name == null) {throw new IllegalArgumentException("Target name cannot be null");}

		String target = name.trim();
		int square = target.indexOf('[');
		if (square >= 0)
		{
			target = target.substring(0, square).trim();
		}
		if (target.length() == 0) {throw new IllegalArgumentException("Target name cannot be empty");}

		int pound = target.indexOf('#');
		if (pound >= 0)
		{
			String cls = target.substring(0, pound).trim();
			if (cls.length() == 0) {throw new IllegalArgumentException("MBean class name cannot be empty");}
		}

		int wildPound = target.indexOf("*#");
		if (wildPound >= 0) {throw new IllegalArgumentException("Invalid wildcard, use only '*'");}

		return target.replace('#', '.');
	}

	private ObjectName parseObjectName(String name)
	{
		if (name == null) {throw new IllegalArgumentException("Target name cannot be null");}

		String target = name.trim();
		int open = target.indexOf('[');
		if (open < 0) {return null;}

		int close = target.indexOf(']');
		if (close < 0) {throw new IllegalArgumentException("Missing closing ObjectName bracket");}

		if (target.substring(open + 1, close).trim().length() == 0) {throw new IllegalArgumentException("Empty string in ObjectName brackets");}

		String objectNameString = target.substring(open + 1, close);
		try
		{
			ObjectName objectName = new ObjectName(objectNameString);
			return objectName;
		}
		catch (MalformedObjectNameException x)
		{
			throw new IllegalArgumentException("Invalid ObjectName");
		}
	}

	private ArrayList parseActions(String actions)
	{
		ArrayList list = new ArrayList();
		if (actions != null)
		{
			StringTokenizer tokenizer = new StringTokenizer(actions, ",");
			while (tokenizer.hasMoreTokens())
			{
				String token = tokenizer.nextToken().trim();
				if (token.length() > 0)
				{
					list.add(token);
				}
			}
		}

		// It is very important for hashCode() that the list is sorted, otherwise different hashcodes are returned
		Collections.sort(list);
		return list;
	}

	private int computeHash()
	{
		int hash = getName().hashCode();
		hash ^= getActionsList().hashCode();
		return hash;
	}
}
