/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.BadAttributeValueExpException;
import javax.management.BadBinaryOpValueExpException;
import javax.management.BadStringOperationException;
import javax.management.InstanceAlreadyExistsException;
import javax.management.InstanceNotFoundException;
import javax.management.IntrospectionException;
import javax.management.InvalidApplicationException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistration;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MBeanServerDelegate;
import javax.management.MBeanServerNotification;
import javax.management.MBeanServerPermission;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.NotificationBroadcaster;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ObjectInstance;
import javax.management.ObjectName;
import javax.management.OperationsException;
import javax.management.QueryExp;
import javax.management.ReflectionException;
import javax.management.RuntimeErrorException;
import javax.management.RuntimeOperationsException;

import mx4j.ImplementationException;
import mx4j.util.Utils;
import mx4j.loading.ClassLoaderObjectInputStream;
import mx4j.log.Log;
import mx4j.log.Logger;
import mx4j.server.interceptor.ContextClassLoaderMBeanServerInterceptor;
import mx4j.server.interceptor.MBeanServerInterceptor;
import mx4j.server.interceptor.MBeanServerInterceptorConfigurator;
import mx4j.server.interceptor.NotificationListenerMBeanServerInterceptor;
import mx4j.server.interceptor.SecurityMBeanServerInterceptor;

/**
 * MBeanServer implementation. <p>
 * The MBeanServer accomplishes these roles:
 * <ul>
 * <li> Returns information about the Agent
 * <li> Acts as a repository for MBeans
 * <li> Acts as an invoker, on behalf of the user, on MBeans
 * </ul>
 * <br>
 * The repository function is delegated to instances of {@link MBeanRepository} classes.
 * This class acts as a factory for MBeanRepository instances, that can be controlled via the system property
 * mx4j.mbeanserver.repository set to the qualified name of the implementation class. <p>
 *
 * This class also acts as an invoker on MBeans. The architecture is interceptor-based, that is whenever you call
 * from a client an MBeanServer method that will end up to call the MBean instance, the call is dispatched to
 * the interceptor chain and eventually to the MBean. <br>
 * The interceptors are configurable via the MBean {@link MBeanServerInterceptorConfigurator}.
 * When the call is about to arrive to the MBean instance, the last interceptor dispatches the call depending on
 * the MBean type: if the MBean is a dynamic MBean, the call is dispatched directly; if the MBean is a standard
 * MBean an {@link MBeanInvoker} is delegated to invoke on the MBean instance.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.11 $
 */
public class MBeanServerImpl implements MBeanServer
{
	private String m_defaultDomain;
	private MBeanRepository m_repository;
	private MBeanServerDelegate m_delegate;
	private ObjectName m_delegateName;
	private MBeanIntrospector m_introspector;
	private MBeanServerInterceptorConfigurator m_invoker;
	private long m_notifications;
	private BaseClassLoaderRepository m_classLoaderRepository;

	private static final String[] EMPTY_PARAMS = new String[0];
	private static final Object[] EMPTY_ARGS = new Object[0];
	private static BaseClassLoaderRepository m_staticClassLoaderRepository = new DefaultClassLoaderRepository();

	public MBeanServerImpl(String defaultDomain)
	{
		Logger logger = getLogger();
		if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Creating MBeanServer instance...");}

		SecurityManager sm = System.getSecurityManager();
		if (sm != null)
		{
			if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Checking permission to create MBeanServer...");}
			sm.checkPermission(new MBeanServerPermission("newMBeanServer"));
		}

		if (defaultDomain == null) {defaultDomain = "";}
		m_defaultDomain = defaultDomain;
		if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("MBeanServer default domain is: '" + m_defaultDomain + "'");}

		m_repository = createMBeanRepository();

		m_classLoaderRepository = createClassLoaderRepository();

		m_introspector = new MBeanIntrospector();

		try {m_delegateName = new ObjectName("JMImplementation", "type", "MBeanServerDelegate");}
		catch (MalformedObjectNameException ignored) {}

		m_delegate = new MBeanServerDelegate();

		try
		{
			ObjectName invokerName = new ReservedObjectName(MBeanServerInterceptorConfigurator.OBJECT_NAME);
			m_invoker = new MBeanServerInterceptorConfigurator(this);
			registerMBean(m_invoker, invokerName);
			m_invoker.registerInterceptor(new ContextClassLoaderMBeanServerInterceptor(), new ReservedObjectName("JMImplementation", "interceptor", "contextclassloader"));
			m_invoker.registerInterceptor(new NotificationListenerMBeanServerInterceptor(), new ReservedObjectName("JMImplementation", "interceptor", "notificationwrapper"));
			m_invoker.registerInterceptor(new SecurityMBeanServerInterceptor(), new ReservedObjectName("JMImplementation", "interceptor", "security"));
			m_invoker.start();
		}
		catch (Exception x)
		{
			logger.error("MBeanServerInterceptorConfigurator cannot be registered", x);
			throw new ImplementationException();
		}

		try
		{
			ReservedObjectName delegateName = new ReservedObjectName("JMImplementation", "type", "MBeanServerDelegate");
			registerMBean(m_delegate, delegateName);
		}
		catch (Exception x)
		{
			logger.error("MBeanServerDelegate cannot be registered", x);
			throw new ImplementationException();
		}

		if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("MBeanServer instance created successfully");}
	}

	public static BaseClassLoaderRepository getStaticClassLoaderRepository()
	{
		return m_staticClassLoaderRepository;
	}

	public BaseClassLoaderRepository getClassLoaderRepository()
	{
		return m_classLoaderRepository;
	}

	private MBeanServerInterceptor getHeadInterceptor()
	{
		synchronized (m_invoker)
		{
			MBeanServerInterceptor interceptor = null;
			while ((interceptor = m_invoker.getHeadInterceptor()) == null)
			{
				// Suspend incoming calls until the user has finished to setup the invoker
				try	{m_invoker.wait();}
				catch (InterruptedException x)
				{
					// Break and return null, will throw NPE somewhere else
					break;
				}
			}
			return interceptor;
		}
	}

	private Logger getLogger()
	{
		return Log.getLogger("mx4j.mbeanserver.implementation");
	}

	private MBeanRepository createMBeanRepository()
	{
		Logger logger = getLogger();

		final String property = "mx4j.mbeanserver.repository";
		if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Checking for system property " + property);}

		String value = (String)AccessController.doPrivileged(new PrivilegedAction()
		{
			public Object run()
			{
				return System.getProperty(property);
			}
		});

		if (value != null)
		{
			if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Property found for custom MBeanServer registry; class is: " + value);}

			try
			{
				MBeanRepository registry = (MBeanRepository)Thread.currentThread().getContextClassLoader().loadClass(value).newInstance();
				if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Custom MBeanServer registry created successfully");}
				return registry;
			}
			catch (Exception x)
			{
				if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Custom MBeanServer registry could not be created", x);}
			}
		}

		return new DefaultMBeanRepository();
	}

	private BaseClassLoaderRepository createClassLoaderRepository()
	{
		Logger logger = getLogger();

		final String property = "mx4j.mbeanserver.classloader.repository";
		if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Checking for system property " + property);}

		String value = (String)AccessController.doPrivileged(new PrivilegedAction()
		{
			public Object run()
			{
				return System.getProperty(property);
			}
		});

		if (value != null)
		{
			if (logger.isEnabledFor(Logger.DEBUG)) {logger.debug("Property found for custom ClassLoaderRepository; class is: " + value);}

			try
			{
				BaseClassLoaderRepository repository = (BaseClassLoaderRepository)Thread.currentThread().getContextClassLoader().loadClass(value).newInstance();
				if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Custom ClassLoaderRepository created successfully");}
				return repository;
			}
			catch (Exception x)
			{
				if (logger.isEnabledFor(Logger.TRACE)) {logger.trace("Custom ClassLoaderRepository could not be created", x);}
			}
		}

		return new DefaultClassLoaderRepository();
	}

	private MBeanRepository getMBeanRepository()
	{
		return m_repository;
	}

	private MBeanMetaData findMBeanMetaData(ObjectName objectName) throws InstanceNotFoundException
	{
		MBeanMetaData metadata = null;
		if (objectName != null)
		{
         objectName = normalizeObjectName(objectName);

			MBeanRepository repository = getMBeanRepository();
			synchronized (repository)
			{
				metadata = repository.get(objectName);
			}
		}
		if (metadata == null) {throw new InstanceNotFoundException("MBeanServer cannot find MBean with ObjectName " + objectName);}
		return metadata;
	}

	public void addNotificationListener(ObjectName observed, ObjectName listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException
	{
		Object mbean = findMBeanMetaData(listener).mbean;
		if (!(mbean instanceof NotificationListener))
		{
			throw new InstanceNotFoundException("ObjectName " + listener + " is not a NotificationListener");
		}
		addNotificationListener(observed, (NotificationListener)mbean, filter, handback);
	}

	public void addNotificationListener(ObjectName observed, NotificationListener listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException
	{
		if (listener == null) {throw new RuntimeOperationsException(new IllegalArgumentException("NotificationListener cannot be null"));}

		MBeanMetaData metadata = findMBeanMetaData(observed);
		Object mbean = metadata.mbean;

		if (!(mbean instanceof NotificationBroadcaster))
		{
			throw new InstanceNotFoundException("ObjectName " + observed + " is not a NotificationBroadcaster");
		}

		addNotificationListenerImpl(metadata, listener, filter, handback);
	}

	private void addNotificationListenerImpl(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException
	{
		getHeadInterceptor().addNotificationListener(metadata, listener, filter, handback);
	}

	public void removeNotificationListener(ObjectName observed, ObjectName listener)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		Object mbean = findMBeanMetaData(listener).mbean;
		if (!(mbean instanceof NotificationListener))
		{
			throw new InstanceNotFoundException("ObjectName " + listener + " is not a NotificationListener");
		}
		removeNotificationListener(observed, (NotificationListener)mbean);
	}

	public void removeNotificationListener(ObjectName observed, NotificationListener listener)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		if (listener == null) {throw new ListenerNotFoundException("NotificationListener cannot be null");}

		MBeanMetaData metadata = findMBeanMetaData(observed);
		Object mbean = metadata.mbean;

		if (!(mbean instanceof NotificationBroadcaster))
		{
			throw new InstanceNotFoundException("ObjectName " + observed + " is not a NotificationBroadcaster");
		}

		removeNotificationListenerImpl(metadata, listener, null, null);
	}

	// Not present in the specification but needed
	public void removeNotificationListener(ObjectName observed, ObjectName listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		Object mbean = findMBeanMetaData(listener).mbean;
		if (!(mbean instanceof NotificationListener))
		{
			throw new InstanceNotFoundException("ObjectName " + listener + " is not a NotificationListener");
		}
		removeNotificationListener(observed, (NotificationListener)mbean, filter, handback);
	}

	// Not present in the specification but needed
	public void removeNotificationListener(ObjectName observed, NotificationListener listener, NotificationFilter filter, Object handback)
			throws InstanceNotFoundException, ListenerNotFoundException
	{
		if (listener == null) {throw new ListenerNotFoundException("NotificationListener cannot be null");}

		MBeanMetaData metadata = findMBeanMetaData(observed);
		Object mbean = metadata.mbean;

		if (!(mbean instanceof NotificationBroadcaster))
		{
			throw new InstanceNotFoundException("ObjectName " + observed + " is not a NotificationBroadcaster");
		}

		removeNotificationListenerImpl(metadata, listener, filter, handback);
	}

	private void removeNotificationListenerImpl(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
			throws ListenerNotFoundException
	{
		getHeadInterceptor().removeNotificationListener(metadata, listener, filter, handback);
	}

	public Object instantiate(String className)
			throws ReflectionException, MBeanException
	{
		return instantiate(className, null, null);
	}

	public Object instantiate(String className, Object[] args, String[] parameters)
			throws ReflectionException, MBeanException
	{
		if (className == null || className.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Class name cannot be null or empty"));}

		try
		{
			Class cls = getClassLoaderRepository().loadClass(className);
			return instantiateImpl(className, cls.getClassLoader(), parameters, args).mbean;
		}
		catch (ClassNotFoundException x)
		{
			throw new ReflectionException(x);
		}
	}

	public Object instantiate(String className, ObjectName loaderName)
			throws ReflectionException, MBeanException, InstanceNotFoundException
	{
		return instantiate(className, loaderName, null, null);
	}

	public Object instantiate(String className, ObjectName loaderName, Object[] args, String[] parameters)
			throws ReflectionException, MBeanException, InstanceNotFoundException
	{
		if (className == null || className.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Class name cannot be null or empty"));}
		// loaderName can be null: means using this class' ClassLoader
		if (loaderName != null && loaderName.isPattern()) {throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName for the ClassLoader cannot be a pattern ObjectName"));}

		ClassLoader cl = findClassLoader(loaderName);
		return instantiateImpl(className, cl, parameters, args).mbean;
	}

	private ClassLoader findClassLoader(ObjectName name) throws InstanceNotFoundException
	{
		if (name == null)
		{
			ClassLoader cl = getClass().getClassLoader();
			if (cl != null) {return cl;}

			// If the classloader is null, then the JMX classes are in the boot classpath;
			// every classloader will do the job, because of delegation. Just avoid to return null
			return ClassLoader.getSystemClassLoader();
		}
		else
		{
			Object mbean = findMBeanMetaData(name).mbean;
			if (!(mbean instanceof ClassLoader)) {throw new InstanceNotFoundException(name.toString());}
			return (ClassLoader)mbean;
		}
	}

	private MBeanMetaData instantiateImpl(String className, ClassLoader classloader, String[] params, Object[] args)
			throws ReflectionException, MBeanException
	{
		if (params == null) {params = EMPTY_PARAMS;}
		if (args == null) {args = EMPTY_ARGS;}

		MBeanMetaData metadata = new MBeanMetaData();
		metadata.classloader = classloader;

		getHeadInterceptor().instantiate(metadata, className, params, args);

		return metadata;
	}

	public ObjectInstance createMBean(String className, ObjectName objectName)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException
	{
		return createMBean(className, objectName, null, null);
	}

	public ObjectInstance createMBean(String className, ObjectName objectName, Object[] args, String[] parameters)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException
	{
		try
		{
			Class cls = getClassLoaderRepository().loadClass(className);

			MBeanMetaData metadata = instantiateImpl(className, cls.getClassLoader(), parameters, args);
//			objectName = normalizeObjectName(objectName);
         metadata.name = objectName;

			registerImpl(metadata);

			return metadata.instance;
		}
		catch (ClassNotFoundException x)
		{
			throw new ReflectionException(x);
		}
	}

	public ObjectInstance createMBean(String className, ObjectName objectName, ObjectName loaderName)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException, InstanceNotFoundException
	{
		return createMBean(className, objectName, loaderName, null, null);
	}

	public ObjectInstance createMBean(String className, ObjectName objectName, ObjectName loaderName, Object[] args, String[] parameters)
			throws ReflectionException, InstanceAlreadyExistsException, MBeanRegistrationException, MBeanException, NotCompliantMBeanException, InstanceNotFoundException
	{
		ClassLoader cl = findClassLoader(loaderName);

		MBeanMetaData metadata = instantiateImpl(className, cl, parameters, args);
//      objectName = normalizeObjectName(objectName);
		metadata.name = objectName;

		registerImpl(metadata);

		return metadata.instance;
	}

	public ObjectInstance registerMBean(Object mbean, ObjectName objectName)
			throws InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException
	{
		if (mbean == null) {throw new RuntimeOperationsException(new IllegalArgumentException("MBean cannot be null"));}

		MBeanMetaData metadata = new MBeanMetaData();
		metadata.mbean = mbean;
		ClassLoader cl = mbean.getClass().getClassLoader();
		if (cl == null)
		{
			// MBean class is in boot classpath, use system classloader
			// For delegation, it will be found anyway,
			// I just don't want to set it to null
			cl = ClassLoader.getSystemClassLoader();
		}
		metadata.classloader = cl;
//      objectName = normalizeObjectName(objectName);
		metadata.name = objectName;

		registerImpl(metadata);

		return metadata.instance;
	}

	private void registerImpl(MBeanMetaData metadata) throws InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException
	{
		m_introspector.introspect(metadata);

		if (!m_introspector.isMBeanCompliant(metadata)) {throw new NotCompliantMBeanException("MBean is not compliant");}

		boolean invokeRegistration = metadata.mbean instanceof MBeanRegistration;

		synchronized (getMBeanRepository())
		{
			if (invokeRegistration)
			{
				getHeadInterceptor().registration(metadata, MBeanServerInterceptor.PRE_REGISTER);
			}

			try
			{
				register(metadata);

				if (invokeRegistration)
				{
					getHeadInterceptor().registration(metadata, MBeanServerInterceptor.POST_REGISTER_TRUE);
				}
			}
			catch (Throwable x)
			{
				if (invokeRegistration)
				{
					try
					{
						getHeadInterceptor().registration(metadata, MBeanServerInterceptor.POST_REGISTER_FALSE);
					}
					catch (MBeanRegistrationException ignored) {/* Ignore this one to rethrow this one */}
				}

				if (x instanceof InstanceAlreadyExistsException) {throw (InstanceAlreadyExistsException)x;}
				else if (x instanceof MBeanRegistrationException) {throw (MBeanRegistrationException)x;}
				else if (x instanceof Exception) {throw new MBeanRegistrationException((Exception)x);}
				else if (x instanceof Error) {throw new MBeanRegistrationException(new RuntimeErrorException((Error)x));}
				else {throw new ImplementationException();}
			}

			if (metadata.mbean instanceof ClassLoader)
			{
				getClassLoaderRepository().addClassLoader((ClassLoader)metadata.mbean);
				getStaticClassLoaderRepository().addClassLoader((ClassLoader)metadata.mbean);
			}

			metadata.instance = new ObjectInstance(metadata.name, metadata.info.getClassName());
		}
	}

	private void register(MBeanMetaData metadata) throws InstanceAlreadyExistsException
	{
      metadata.name = normalizeObjectName(metadata.name);
		ObjectName objectName = metadata.name;
		if (objectName == null || objectName.isPattern()) {throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName cannot be null or a pattern ObjectName"));}
		if (objectName.getDomain().equals("JMImplementation") && !(objectName instanceof ReservedObjectName)) {throw new RuntimeOperationsException(new IllegalArgumentException("Domain 'JMImplementation' is reserved for the JMX Agent"));}

		MBeanRepository repository = getMBeanRepository();
		synchronized (repository)
		{
			if (repository.get(objectName) != null) {throw new InstanceAlreadyExistsException(objectName.toString());}
			repository.put(objectName, metadata);
		}

		notify(objectName, MBeanServerNotification.REGISTRATION_NOTIFICATION);
	}

	private void notify(ObjectName objectName, String notificationType)
	{
		long sequenceNumber = 0;
		synchronized (MBeanServerImpl.class)
		{
			sequenceNumber = m_notifications;
			++m_notifications;
		}

		m_delegate.sendNotification(new MBeanServerNotification(notificationType, m_delegateName, sequenceNumber, objectName));
	}

	public void unregisterMBean(ObjectName objectName)
			throws InstanceNotFoundException, MBeanRegistrationException
	{
		if (objectName == null || objectName.isPattern()) {throw new RuntimeOperationsException(new IllegalArgumentException("ObjectName cannot be null or a pattern ObjectName"));}
		if (objectName.getDomain().equals("JMImplementation")) {throw new MBeanRegistrationException(new RuntimeOperationsException(new IllegalArgumentException("Domain 'JMImplementation' is reserved for the JMX Agent")));}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		try
		{
			boolean invokeRegistration = metadata.mbean instanceof MBeanRegistration;

			if (invokeRegistration)
			{
				getHeadInterceptor().registration(metadata, MBeanServerInterceptor.PRE_DEREGISTER);
			}

			unregister(metadata);

			if (invokeRegistration)
			{
				getHeadInterceptor().registration(metadata, MBeanServerInterceptor.POST_DEREGISTER);
			}

			if (metadata.mbean instanceof ClassLoader)
			{
				getClassLoaderRepository().removeClassLoader((ClassLoader)metadata.mbean);
				getStaticClassLoaderRepository().removeClassLoader((ClassLoader)metadata.mbean);
			}

			// PENDING: other things should be done ? for example, removing all listeners if it's a broadcaster ?
		}
		catch (MBeanRegistrationException x) {throw x;}
		catch (Exception x) {throw new MBeanRegistrationException(x);}
		catch (Error x) {throw new MBeanRegistrationException(new RuntimeErrorException(x));}
	}

	private void unregister(MBeanMetaData metadata)
	{
		ObjectName objectName = metadata.name;

		MBeanRepository repository = getMBeanRepository();
		synchronized (repository)
		{
			repository.remove(objectName);
		}

		notify(objectName, MBeanServerNotification.UNREGISTRATION_NOTIFICATION);
	}

	public ObjectInputStream deserialize(String className, ObjectName loaderName, byte[] bytes)
			throws InstanceNotFoundException, OperationsException, ReflectionException
	{
		ClassLoader cl = findClassLoader(loaderName);
		return deserializeImpl(cl, bytes);
	}

	public ObjectInputStream deserialize(String className, byte[] bytes)
			throws OperationsException, ReflectionException
	{
		if (className == null || className.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid class name"));}

		// Find the classloader that loaded the given className
		Class cls = null;
		try
		{
			cls = getClass().getClassLoader().loadClass(className);
		}
		catch (ClassNotFoundException x)
		{
			// Try the ClassLoaderRepository
			try
			{
				cls = getClassLoaderRepository().loadClass(className);
			}
			catch (ClassNotFoundException xx)
			{
				throw new ReflectionException(xx);
			}
		}

		return deserializeImpl(cls.getClassLoader(), bytes);
	}

	public ObjectInputStream deserialize(ObjectName objectName, byte[] bytes)
			throws InstanceNotFoundException, OperationsException
	{
		MBeanMetaData metadata = findMBeanMetaData(objectName);
		return deserializeImpl(metadata.classloader, bytes);
	}

	private ObjectInputStream deserializeImpl(ClassLoader classloader, byte[] bytes) throws OperationsException
	{
		if (bytes == null || bytes.length == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid bytes"));}

      ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
		try {return new ClassLoaderObjectInputStream(bais, classloader);}
		catch (IOException x) {throw new OperationsException(x.toString());}
	}

	public Object getAttribute(ObjectName objectName, String attribute)
			throws InstanceNotFoundException, MBeanException, AttributeNotFoundException, ReflectionException
	{
		if (attribute == null || attribute.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute"));}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		return getHeadInterceptor().getAttribute(metadata, attribute);
	}

	public void setAttribute(ObjectName objectName, Attribute attribute)
			throws InstanceNotFoundException, AttributeNotFoundException, InvalidAttributeValueException, MBeanException, ReflectionException
	{
		if (attribute == null || attribute.getName().trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute"));}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		getHeadInterceptor().setAttribute(metadata, attribute);
	}

	public AttributeList getAttributes(ObjectName objectName, String[] attributes)
			throws InstanceNotFoundException, ReflectionException
	{
		if (attributes == null || attributes.length == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute list"));}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		return getHeadInterceptor().getAttributes(metadata, attributes);
	}

	public AttributeList setAttributes(ObjectName objectName, AttributeList attributes)
			throws InstanceNotFoundException, ReflectionException
	{
		if (attributes == null) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid attribute list"));}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		return getHeadInterceptor().setAttributes(metadata, attributes);
	}

	public Object invoke(ObjectName objectName, String methodName, Object[] args, String[] parameters)
			throws InstanceNotFoundException, MBeanException, ReflectionException
	{
		if (methodName == null || methodName.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid operation name"));}
		if (args == null) {args = EMPTY_ARGS;}
		if (parameters == null) {parameters = EMPTY_PARAMS;}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		return getHeadInterceptor().invoke(metadata, methodName, parameters, args);
	}

	public String getDefaultDomain()
	{
		return m_defaultDomain;
	}

	public Integer getMBeanCount()
	{
		MBeanRepository repository = getMBeanRepository();
		synchronized (repository)
		{
			return new Integer(repository.size());
		}
	}

	public boolean isRegistered(ObjectName objectName)
	{
		try
		{
			return findMBeanMetaData(objectName) != null;
		}
		catch (InstanceNotFoundException x)
		{
			return false;
		}
	}

	public MBeanInfo getMBeanInfo(ObjectName objectName)
			throws InstanceNotFoundException, IntrospectionException, ReflectionException
	{
		MBeanMetaData metadata = findMBeanMetaData(objectName);

		if (metadata.dynamic)
		{
			MBeanInfo info = getHeadInterceptor().getMBeanInfo(metadata);
			metadata.info = info;
		}

		return (MBeanInfo)metadata.info.clone();
	}

	public ObjectInstance getObjectInstance(ObjectName objectName)
			throws InstanceNotFoundException
	{
		MBeanMetaData metadata = findMBeanMetaData(objectName);
		return metadata.instance;
	}

	public boolean isInstanceOf(ObjectName objectName, String className)
			throws InstanceNotFoundException
	{
		if (className == null || className.trim().length() == 0) {throw new RuntimeOperationsException(new IllegalArgumentException("Invalid class name"));}

		MBeanMetaData metadata = findMBeanMetaData(objectName);

		try
		{
			Class cls = metadata.classloader.loadClass(className);

			return cls.isInstance(metadata.mbean);
		}
		catch (ClassNotFoundException x)
		{
			return false;
		}
	}

	public Set queryMBeans(ObjectName patternName, QueryExp filter)
	{
		Set match = queryNames(patternName, filter);

		Set set = new HashSet();
		for (Iterator i = match.iterator(); i.hasNext();)
		{
			ObjectName name = (ObjectName)i.next();
			try
			{
				MBeanMetaData metadata = findMBeanMetaData(name);
				set.add(metadata.instance);
			}
			catch (InstanceNotFoundException ignored)
			{
				// A concurrent thread removed the MBean after queryNames, ignore
			}
		}
		return set;
	}

	public Set queryNames(ObjectName patternName, QueryExp filter)
	{
		// First retrieve the scope of the query: all mbeans matching the patternName
		Set scope = findMBeansByPattern(patternName);

		// Second filter the scope using the given filter
		Set match = filterMBeans(scope, filter);

		return match;
	}

	private Set findMBeansByPattern(ObjectName pattern)
	{
		if (pattern == null)
		{
			try {pattern = new ObjectName("*:*");}
			catch (MalformedObjectNameException ignored) {}
		}

      pattern = normalizeObjectName(pattern);

		String patternDomain = pattern.getDomain();
		Hashtable patternProps = pattern.getKeyPropertyList();

		Set set = new HashSet();
		MBeanRepository repository = getMBeanRepository();
		synchronized (repository)
		{
			for (Iterator i = repository.iterator(); i.hasNext();)
			{
				MBeanMetaData metadata = (MBeanMetaData)i.next();
				ObjectName name = metadata.name;
				Hashtable props = name.getKeyPropertyList();

				String domain = name.getDomain();
				if (Utils.wildcardMatch(patternDomain, domain))
				{
					// Domain matches, now check properties
					if (pattern.isPropertyPattern())
					{
						// A property pattern with no entries, can only be '*'
						if (patternProps.size() == 0)
						{
							// User wants all properties
							set.add(name);
						}
						else
						{
							// Loop on the properties of the pattern.
							// If one is not found then the current ObjectName does not match
							boolean found = true;
							for (Iterator j = patternProps.entrySet().iterator(); j.hasNext();)
							{
								Map.Entry entry = (Map.Entry)j.next();
								Object patternKey = entry.getKey();
								Object patternValue = entry.getValue();
								if (patternKey.equals("*")) {continue;}

								// Try to see if the current ObjectName contains this entry
								if (!props.containsKey(patternKey))
								{
									// Not even the key is present
									found = false;
									break;
								}
								else
								{
									// The key is present, let's check if the values are equal
									Object value = props.get(patternKey);
									if (value == null && patternValue == null)
									{
										// Values are equal, go on with next pattern entry
										continue;
									}
									if (value != null && value.equals(patternValue))
									{
										// Values are equal, go on with next pattern entry
										continue;
									}
									// Here values are different
									found = false;
									break;
								}
							}
							if (found)
							{
								set.add(name);
							}
						}
					}
					else
					{
						if (props.entrySet().equals(patternProps.entrySet()))
						{
							set.add(name);
						}
					}
				}
			}
		}
		return set;
	}

	private Set filterMBeans(Set scope, QueryExp filter)
	{
		if (filter == null) {return scope;}

		Set set = new HashSet();
		for (Iterator i = scope.iterator(); i.hasNext();)
		{
			ObjectName name = (ObjectName)i.next();
			filter.setMBeanServer(this);
			try
			{
				if (filter.apply(name)) {set.add(name);}
			}
			catch (BadStringOperationException ignored) {}
			catch (BadBinaryOpValueExpException ignored) {}
			catch (BadAttributeValueExpException x) {}
			catch (InvalidApplicationException x) {}
			catch (SecurityException x) {}
		}
		return set;
	}

   private ObjectName normalizeObjectName(ObjectName name)
   {
      if (name == null) {return null;}

      String defaultDomain = getDefaultDomain();
      String domain = name.getDomain();

      if (domain.length() == 0 && defaultDomain.length() > 0)
      {
         // The given object name specifies the abbreviated form to indicate the default domain,
         // ie ':key=value', the empty string as domain. I must convert this abbreviated form
         // to the full one, if the default domain of this mbeanserver is not the empty string as well
         StringBuffer buffer = new StringBuffer(getDefaultDomain()).append(":").append(name.getKeyPropertyListString());
         if (name.isPropertyPattern())
         {
            if (name.getKeyPropertyList().size() > 0)
            {
               buffer.append(",*");
            }
            else
            {
               buffer.append("*");
            }
         }
         try {name = new ObjectName(buffer.toString());}
         catch (MalformedObjectNameException ignored) {}
      }
      return name;
   }

	private static class ReservedObjectName extends ObjectName
	{
		private ReservedObjectName(String name) throws MalformedObjectNameException {super(name);}
      private ReservedObjectName(String domain, String key, String value) throws MalformedObjectNameException {super(domain, key, value);}
	}
}
