/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server.interceptor;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.InvalidAttributeValueException;
import javax.management.ListenerNotFoundException;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanRegistrationException;
import javax.management.NotificationFilter;
import javax.management.NotificationListener;
import javax.management.ReflectionException;

import mx4j.server.MBeanMetaData;


/**
 * This interceptor sets the context class loader to the proper value for incoming calls.
 * It saves the current context class loader, set the context class loader to be the MBean's class loader for
 * the current call, and on return re-set the context class loader to the previous value
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.3 $
 */
public class ContextClassLoaderMBeanServerInterceptor extends DefaultMBeanServerInterceptor
{
	public String getType()
	{
		return "contextclassloader";
	}

	public void addNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback)
	{
		if (!isEnabled()) {super.addNotificationListener(metadata, listener, filter, handback);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			super.addNotificationListener(metadata, listener, filter, handback);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public void removeNotificationListener(MBeanMetaData metadata, NotificationListener listener, NotificationFilter filter, Object handback) throws ListenerNotFoundException
	{
		if (!isEnabled()) {super.removeNotificationListener(metadata, listener, filter, handback);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			super.removeNotificationListener(metadata, listener, filter, handback);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public void instantiate(MBeanMetaData metadata, String className, String[] params, Object[] args) throws ReflectionException, MBeanException
	{
		if (!isEnabled()) {super.instantiate(metadata, className, params, args);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			super.instantiate(metadata, className, params, args);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public void registration(MBeanMetaData metadata, int operation) throws MBeanRegistrationException
	{
		if (!isEnabled()) {super.registration(metadata, operation);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			super.registration(metadata, operation);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public MBeanInfo getMBeanInfo(MBeanMetaData metadata)
	{
		if (!isEnabled()) {return super.getMBeanInfo(metadata);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			return super.getMBeanInfo(metadata);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public Object invoke(MBeanMetaData metadata, String method, String[] params, Object[] args) throws MBeanException, ReflectionException
	{
		if (!isEnabled()) {return super.invoke(metadata, method, params, args);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			return super.invoke(metadata, method, params, args);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public AttributeList getAttributes(MBeanMetaData metadata, String[] attributes)
	{
		if (!isEnabled()) {return super.getAttributes(metadata, attributes);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			return super.getAttributes(metadata, attributes);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public AttributeList setAttributes(MBeanMetaData metadata, AttributeList attributes)
	{
		if (!isEnabled()) {return super.setAttributes(metadata, attributes);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			return super.setAttributes(metadata, attributes);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public Object getAttribute(MBeanMetaData metadata, String attribute) throws MBeanException, AttributeNotFoundException, ReflectionException
	{
		if (!isEnabled()) {return super.getAttribute(metadata, attribute);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			return super.getAttribute(metadata, attribute);
		}
		finally {current.setContextClassLoader(cl);}
	}

	public void setAttribute(MBeanMetaData metadata, Attribute attribute) throws MBeanException, AttributeNotFoundException, InvalidAttributeValueException, ReflectionException
	{
		if (!isEnabled()) {super.setAttribute(metadata, attribute);}

		Thread current = Thread.currentThread();
		ClassLoader cl = current.getContextClassLoader();
		try
		{
			current.setContextClassLoader(metadata.classloader);
			super.setAttribute(metadata, attribute);
		}
		finally {current.setContextClassLoader(cl);}
	}
}
