/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.server.interceptor;

import java.util.ArrayList;

import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.ObjectName;

/**
 * MBean that configures the MBeanServer --> MBean interceptor chain.
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.2 $
 */
public class MBeanServerInterceptorConfigurator implements MBeanServerInterceptorConfiguratorMBean
{
	private MBeanServer m_server;
	private MBeanServerInterceptor m_last;
	private MBeanServerInterceptor m_first;
	private ArrayList m_interceptors = new ArrayList();
	private boolean m_running;

	public MBeanServerInterceptorConfigurator(MBeanServer server)
	{
		m_server = server;

		// This interceptor must be the last one
		m_last = new InvokerMBeanServerInterceptor(m_server);
	}

	/**
	 * Append the given interceptor to the interceptor chain.
	 * This method has effect only if this MBean is not running.
	 * The interceptor is not registered in the MBeanServer; use {@link #registerInterceptor} to do so.
	 * @see #isRunning
	 */
	public synchronized void addInterceptor(MBeanServerInterceptor interceptor)
	{
		if (!isRunning())
		{
			m_interceptors.add(interceptor);
		}
	}

	/**
	 * Append the given interceptor to the interceptor chain.
	 * This method has effect only if this MBean is not running.
	 * The interceptor is registered in the MBeanServer; use {@link #addInterceptor} to avoid so.
	 * @see #isRunning
	 */
	public void registerInterceptor(MBeanServerInterceptor interceptor, ObjectName name) throws MBeanRegistrationException
	{
		if (!isRunning())
		{
			try
			{
				m_server.registerMBean(interceptor, name);
				addInterceptor(interceptor);
			}
			catch (MBeanRegistrationException x)
			{
				throw x;
			}
			catch (Exception x)
			{
				throw new MBeanRegistrationException(x);
			}
		}
	}

	/**
	 * Returns the head interceptor.
	 * There always is a head interceptor.
	 */
	public synchronized MBeanServerInterceptor getHeadInterceptor()
	{
		// Sync'ed because m_first can be modified and read by 2 different threads
		return m_first;
	}

	public synchronized void start()
	{
		if (!isRunning())
		{
			if (m_interceptors.size() > 0)
			{
				for (int i = 1; i < m_interceptors.size(); ++i)
				{
					MBeanServerInterceptor previous = (MBeanServerInterceptor)m_interceptors.get(i - 1);
					MBeanServerInterceptor next = (MBeanServerInterceptor)m_interceptors.get(i);
					previous.setNext(next);
				}
				MBeanServerInterceptor interceptor = (MBeanServerInterceptor)m_interceptors.get(m_interceptors.size() - 1);
				interceptor.setNext(m_last);
				m_first = (MBeanServerInterceptor)m_interceptors.get(0);
			}
			else
			{
				m_first = m_last;
			}

			m_running = true;

			// See usages of getHeadInterceptor()
			notifyAll();
		}
	}

	public synchronized void stop()
	{
		if (isRunning())
		{
			m_first = null;
		}
	}

	public synchronized boolean isRunning()
	{
		return m_running;
	}

	public synchronized void clearInterceptors()
	{
		if (!isRunning())
		{
			m_interceptors.clear();
		}
	}
}
