/*
 * Copyright (C) MX4J.
 * All rights reserved.
 *
 * This software is distributed under the terms of the MX4J License version 1.0.
 * See the terms of the MX4J License in the documentation provided with this software.
 */

package mx4j.util;

import java.lang.reflect.Proxy;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.List;

import javax.management.ObjectName;
import javax.management.MBeanServer;
import javax.management.MBeanServerFactory;
import javax.management.InstanceNotFoundException;
import javax.management.MBeanException;
import javax.management.ReflectionException;
import javax.management.RuntimeOperationsException;
import javax.management.RuntimeMBeanException;
import javax.management.RuntimeErrorException;
import javax.management.Attribute;

/**
 * A proxy for invocation on standard MBeans (does not work for DynamicMBeans). <p>
 * Usage example:
 * <pre>
 * public interface MyServiceMBean {...}
 * public class MyService implements MyServiceMBean {...}
 * pulic class Main
 * {
 *    public static void main(String[] args) throws Exception
 *    {
 *       MBeanServer server = ...;
 *       ObjectName myServiceObjectName = ...;
 *
 *       MyServiceMBean mbean = (MyServiceMBean)StandardMBeanProxy.create(MyServiceMBean.class, server, myServiceObjectName);
 *
 *       ...
 *    }
 * }
 * </pre>
 *
 * @author <a href="mailto:biorn_steedom@users.sourceforge.net">Simone Bordet</a>
 * @version $Revision: 1.3 $
 */
public class StandardMBeanProxy
{
   /**
    * Creates a proxy with the given MBean interface for an MBean with the specified name
    * living in the MBeanServer returned by <code>MBeanServerFactory.findMBeanServer(null).get(0)</code>.
    */
	public static Object create(Class mbeanInterface, ObjectName name)
	{
		return create(mbeanInterface, null, name);
	}

   /**
    * Creates a proxy with the given MBean interface for an MBean with the specified name
    * living in the specified MBeanServer.
    */
	public static Object create(Class mbeanInterface, MBeanServer server, ObjectName name)
	{
		if (mbeanInterface == null) {throw new IllegalArgumentException("MBean interface cannot be null");}
		if (!mbeanInterface.isInterface()) {throw new IllegalArgumentException("Class parameter must be an interface");}
		if (name == null) {throw new IllegalArgumentException("MBean name cannot be null");}

		if (server == null)
		{
			List list = MBeanServerFactory.findMBeanServer(null);
			if (list.size() > 0)
			{
				server = (MBeanServer)list.get(0);
			}
			else {throw new IllegalStateException("Cannot find MBeanServer");}
		}

		// The client must be able to cast the returned object to the mbeanInterface it passes,
		// so the classloader must be the same
		ClassLoader loader = mbeanInterface.getClassLoader();
		return Proxy.newProxyInstance(loader, new Class[] {mbeanInterface}, new Handler(server, name));
	}

	private static class Handler implements InvocationHandler
	{
		private MBeanServer m_server;
		private ObjectName m_name;

		private Handler(MBeanServer server, ObjectName name)
		{
			m_server = server;
			m_name = name;
		}

		public Object invoke(Object proxy, Method method, Object[] args) throws Throwable
		{
			// Needed for no-parameter methods
			if (args == null) {args = new Object[0];}

			// No need to check for consistency between the signature and the args parameter,
			// since the invocation it is not done by reflection, but statically

			if (Utils.isAttributeSetter(method))
			{
				String name = method.getName().substring(3);
				Attribute attribute = new Attribute(name, args[0]);
				try
				{
					m_server.setAttribute(m_name, attribute);
					return null;
				}
				catch (MBeanException x)
				{
					throw x.getTargetException();
				}
				catch (ReflectionException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeOperationsException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeMBeanException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeErrorException x)
				{
					throw x.getTargetError();
				}
			}
			else if (Utils.isAttributeGetter(method))
			{
				String n = method.getName();
				String name = null;
				if (n.startsWith("is")) {name = n.substring(2);}
				else {name = n.substring(3);}
				try
				{
					return m_server.getAttribute(m_name, name);
				}
				catch (MBeanException x)
				{
					throw x.getTargetException();
				}
				catch (ReflectionException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeOperationsException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeMBeanException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeErrorException x)
				{
					throw x.getTargetError();
				}
			}
			else
			{
				Class[] parameters = method.getParameterTypes();
				String[] params = new String[parameters.length];
				for (int i = 0; i < parameters.length; ++i)
				{
					params[i] = parameters[i].getName();
				}

				try
				{
					return m_server.invoke(m_name, method.getName(), args, params);
				}
				catch (MBeanException x)
				{
					throw x.getTargetException();
				}
				catch (ReflectionException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeOperationsException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeMBeanException x)
				{
					throw x.getTargetException();
				}
				catch (RuntimeErrorException x)
				{
					throw x.getTargetError();
				}
			}
		}
	}
}
